"""
Handle local builtin datasets
"""

import os
from .base import  get_list_of_files



dirs = [
    "10740", "arcgis", "baltim", "berlin", "book", "burkitt", "calemp",
    "chicago", "columbus", "desmith", "geodanet", "georgia",
    "juvenile", "Line", "mexico", "networks", "Point", "Polygon",
    "Polygon_Holes", "sids2", "snow_maps", "stl", "street_net_pts", "tests",
    "tokyo", "us_income", "virginia", "wmat"
]


class LocalExample:
    """
    Builtin pysal example dataset
    """
    def __init__(self, name, dirname):
        self.name = name
        self.dirname = dirname
        self.installed = True
        self.description = self.get_description()

    def get_file_list(self):
        return get_list_of_files(self.dirname)

    def get_path(self, file_name, verbose=True):
        """
        get path for local file
        """
        file_list = self.get_file_list()
        for file_path in file_list:
            base_name = os.path.basename(file_path)
            if file_name == base_name:
                return file_path
        if verbose:
            print("{} is not a file in this example".format(file_name))
        return None

    def explain(self):
        """
        Provide a description of the example
        """
        description = [f for f in self.get_file_list() if "README.md" in f][0]
        with open(description, 'r', encoding="utf8") as f:
            print(f.read())

    def get_description(self):
        description = [f for f in self.get_file_list() if "README.md" in f][0]
        with open(description, 'r', encoding="utf8") as f:
            lines = f.readlines()
        return lines[3].strip()




builtin_root = os.path.dirname(__file__)

paths = [os.path.join(builtin_root, local) for local in dirs]
paths = zip(dirs, paths)

datasets = {}

for name, pth in paths:
    files = get_list_of_files(pth)
    file_names = [os.path.basename(file) for file in files]
    if "README.md" in file_names:
        example = LocalExample(name, pth)
        datasets[name] = example



import os
import unittest
from .._contW_lists import ContiguityWeightsLists, ROOK, QUEEN
from ..weights import W
from ...io.fileio import FileIO as ps_open


try:
    import geopandas as gpd
except ImportError:
    gpd = None


from ... import examples as pysal_examples



class TestContiguityWeights(unittest.TestCase):
    def setUp(self):
        """ Setup the binning contiguity weights"""
        shpObj = ps_open(pysal_examples.get_path('virginia.shp'), 'r')
        self.binningW = ContiguityWeightsLists(shpObj, QUEEN)
        shpObj.close()

    def test_w_type(self):
        self.assertTrue(isinstance(self.binningW, ContiguityWeightsLists))

    def test_QUEEN(self):
        self.assertEqual(QUEEN, 1)

    def test_ROOK(self):
        self.assertEqual(ROOK, 2)

    def test_ContiguityWeightsLists(self):
        self.assertTrue(hasattr(self.binningW, 'w'))
        self.assertTrue(issubclass(dict, type(self.binningW.w)))
        self.assertEqual(len(self.binningW.w), 136)

    def test_nested_polygons(self):
        # load queen gal file created using Open Geoda.
        geodaW = ps_open(
            pysal_examples.get_path('virginia.gal'), 'r').read()
        # build matching W with pysal
        pysalWb = self.build_W(
            pysal_examples.get_path('virginia.shp'), QUEEN, 'POLY_ID')
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            self.assertEqual(geoda_neighbors, pysalb_neighbors)

    def test_true_rook(self):
        # load queen gal file created using Open Geoda.
        geodaW = ps_open(pysal_examples.get_path('rook31.gal'), 'r').read()
        # build matching W with pysal
        #pysalW = pysal.rook_from_shapefile(pysal_examples.get_path('rook31.shp'),','POLY_ID')
        pysalWb = self.build_W(
            pysal_examples.get_path('rook31.shp'), ROOK, 'POLY_ID')
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            self.assertEqual(geoda_neighbors, pysalb_neighbors)

    def test_true_rook2(self):
        # load queen gal file created using Open Geoda.

        stl = pysal_examples.load_example('stl')
        gal_file = test_file = stl.get_path('stl_hom_rook.gal')
        geodaW = ps_open(gal_file, 'r').read()
        # build matching W with pysal
        pysalWb = self.build_W(stl.get_path(
            'stl_hom.shp'), ROOK, 'POLY_ID_OG')
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            self.assertEqual(geoda_neighbors, pysalb_neighbors)

    def test_true_rook3(self):
        # load queen gal file created using Open Geoda.
        geodaW = ps_open(
            pysal_examples.get_path('virginia_rook.gal'), 'r').read()
        # build matching W with pysal
        pysalWb = self.build_W(
            pysal_examples.get_path('virginia.shp'), ROOK, 'POLY_ID')
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            self.assertEqual(geoda_neighbors, pysalb_neighbors)

    @unittest.skipIf(gpd is None, 'geopandas is missing in the testing environment')
    def test_shapely(self):
        pysalneighbs = ContiguityWeightsLists(ps_open(
            pysal_examples.get_path('virginia.shp')), ROOK)
        gdf = gpd.read_file(pysal_examples.get_path('virginia.shp')) 
        shplyneighbs = ContiguityWeightsLists(gdf.geometry.tolist(), ROOK)
        self.assertEqual(pysalneighbs.w, shplyneighbs.w)
        pysalneighbs = ContiguityWeightsLists(ps_open(
            pysal_examples.get_path('virginia.shp')), QUEEN)
        shplyneighbs = ContiguityWeightsLists(gdf.geometry.tolist(), QUEEN)
        self.assertEqual(pysalneighbs.w, shplyneighbs.w)

    def build_W(self, shapefile, type, idVariable=None):
        """ Building 2 W's the hard way.  We need to do this so we can test both rtree and binning """
        dbname = os.path.splitext(shapefile)[0] + '.dbf'
        db = ps_open(dbname)
        shpObj = ps_open(shapefile)
        neighbor_data = ContiguityWeightsLists(shpObj, type).w
        neighbors = {}
        weights = {}
        if idVariable:
            ids = db.by_col[idVariable]
            self.assertEqual(len(ids), len(set(ids)))
            for key in neighbor_data:
                id = ids[key]
                if id not in neighbors:
                    neighbors[id] = set()
                neighbors[id].update([ids[x] for x in neighbor_data[key]])
            for key in neighbors:
                neighbors[key] = list(neighbors[key])
            binningW = W(neighbors, id_order=ids)
        else:
            neighbors[key] = list(neighbors[key])
            binningW = W(neighbors)
        return binningW

#suite = unittest.TestLoader().loadTestsFromTestCase(_TestContiguityWeights)

if __name__ == '__main__':
    #runner = unittest.TextTestRunner()
    #runner.run(suite)
    unittest.main()



"""
Handle remote datasets
"""
from bs4 import BeautifulSoup
import requests
from .base import (PYSALDATA, Example, get_list_of_files,
                   get_data_home)



# Geoda Center Data Sets

url = "https://geodacenter.github.io/data-and-lab//"
page = requests.get(url)
soup = BeautifulSoup(page.text, 'html.parser')
samples = soup.find(class_='samples')
rows = samples.find_all('tr')
datasets = {}
for row in rows[1:]:
    data = row.find_all('td')
    name = data[0].text.strip()
    description = data[1].text
    n = data[2].text
    k = data[3].text
    targets = row.find_all('a')
    download_url = targets[1].attrs['href']
    explain_url = targets[0].attrs['href']
    datasets[name] = Example(name, description, n, k, download_url, explain_url)


# Other Remotes

# rio
name = 'Rio Grande do Sul'
description = 'Cities of the Brazilian State of Rio Grande do Sul'
n = 497
k = 3
download_url = 'https://github.com/sjsrey/rio_grande_do_sul/archive/master.zip'
explain_url = 'https://raw.githubusercontent.com/sjsrey/rio_grande_do_sul/master/README.md'
datasets[name] = Example(name, description, n, k, download_url, explain_url)

# nyc bikes
name = 'nyc_bikes'
description = 'New York City Bike Trips'
n = 14042
k = 27
download_url = 'https://github.com/sjsrey/nyc_bikes/archive/master.zip'
explain_url = 'https://raw.githubusercontent.com/sjsrey/nyc_bikes/master/README.md'
datasets[name] = Example(name, description, n, k, download_url, explain_url)

# taz
name = 'taz'
description = 'Traffic Analysis Zones in So. California'
n = 4109
k = 14
download_url = 'https://github.com/sjsrey/taz/archive/master.zip'
explain_url = 'https://raw.githubusercontent.com/sjsrey/taz/master/README.md'
datasets[name] = Example(name, description, n, k, download_url, explain_url)

# clearwater
name = 'clearwater'
description = 'mgwr testing dataset'
n = 239
k = 14
download_url = 'https://github.com/sjsrey/clearwater/archive/master.zip'
explain_url = 'https://raw.githubusercontent.com/sjsrey/clearwater/master/README.md'
datasets[name] = Example(name, description, n, k, download_url, explain_url)

# newHaven
name = 'newHaven'
description = 'Network testing dataset'
n = 3293
k = 5
download_url = 'https://github.com/sjsrey/newHaven/archive/master.zip'
explain_url = 'https://raw.githubusercontent.com/sjsrey/newHaven/master/README.md'
datasets[name] = Example(name, description, n, k, download_url, explain_url)


# remove Cars dataset as it is broken
datasets.pop('Cars')

def download(datasets=datasets):
    '''
    Download all known remotes
    '''
    names = list(datasets.keys())
    names.sort()
    for name in names:
        print(name)
        example = datasets[name]
        try:
            example.download()
        except:
            print('Example not downloaded: {}'.format(name))



from .. import contiguity as c
from ..weights import W
from .. import util
from ...common import pandas
from ...io.fileio import FileIO as ps_open
from ...io import geotable as pdio

from ... import examples as pysal_examples
import unittest as ut
import numpy as np

PANDAS_EXTINCT = pandas is None
try:
    import geopandas
    GEOPANDAS_EXTINCT = False
except ImportError:
    GEOPANDAS_EXTINCT = True

class Contiguity_Mixin(object):
    polygon_path = pysal_examples.get_path('columbus.shp')
    point_path = pysal_examples.get_path('baltim.shp')
    f = ps_open(polygon_path) # our file handler
    polygons = f.read() # our iterable
    f.seek(0) #go back to head of file
    cls = object # class constructor
    known_wi = None #index of known w entry to compare
    known_w = dict() #actual w entry
    known_name = known_wi
    known_namedw = known_w
    idVariable = None # id variable from file or column

    def setUp(self):
        self.__dict__.update({k:v for k,v in list(Contiguity_Mixin.__dict__.items())
            if not k.startswith('_')})
    
    def runTest(self):
        pass 

    def test_init(self):
        # basic
        w = self.cls(self.polygons)
        self.assertEqual(w[self.known_wi], self.known_w)

        # sparse
        #w = self.cls(self.polygons, sparse=True)
        #srowvec = ws.sparse[self.known_wi].todense().tolist()[0]
        #this_w = {i:k for i,k in enumerate(srowvec) if k>0}
        #self.assertEqual(this_w, self.known_w)
        #ids = ps.weights2.utils.get_ids(self.polygon_path, self.idVariable)

        # named
        ids = util.get_ids(self.polygon_path, self.idVariable)
        w = self.cls(self.polygons, ids = ids)
        self.assertEqual(w[self.known_name], self.known_namedw)

    def test_from_iterable(self):
        w = self.cls.from_iterable(self.f)
        self.f.seek(0)
        self.assertEqual(w[self.known_wi], self.known_w)
        
    def test_from_shapefile(self):
        # basic
        w = self.cls.from_shapefile(self.polygon_path)
        self.assertEqual(w[self.known_wi], self.known_w)

        # sparse
        ws = self.cls.from_shapefile(self.polygon_path, sparse=True)
        srowvec = ws.sparse[self.known_wi].todense().tolist()[0]
        this_w = {i:k for i,k in enumerate(srowvec) if k>0}
        self.assertEqual(this_w, self.known_w)

        # named
        w = self.cls.from_shapefile(self.polygon_path, idVariable=self.idVariable)
        self.assertEqual(w[self.known_name], self.known_namedw)

    def test_from_array(self):
        # test named, sparse from point array
        pass 

    @ut.skipIf(PANDAS_EXTINCT, 'Missing pandas')
    def test_from_dataframe(self):
        # basic
        df = pdio.read_files(self.polygon_path)
        w = self.cls.from_dataframe(df)
        self.assertEqual(w[self.known_wi], self.known_w)

        # named geometry
        df.rename(columns={'geometry':'the_geom'}, inplace=True)
        w = self.cls.from_dataframe(df, geom_col = 'the_geom')
        self.assertEqual(w[self.known_wi], self.known_w)

        # named geometry + named obs
        w = self.cls.from_dataframe(df, geom_col='the_geom', idVariable=self.idVariable)
        self.assertEqual(w[self.known_name], self.known_namedw)


class Test_Queen(ut.TestCase, Contiguity_Mixin):
    def setUp(self):
        Contiguity_Mixin.setUp(self)
        
        self.known_wi = 4
        self.known_w  = {2: 1.0, 3: 1.0, 5: 1.0, 7: 1.0,
                             8: 1.0, 10: 1.0, 14: 1.0, 15: 1.0}
        self.cls = c.Queen
        self.idVariable = 'POLYID'
        self.known_name = 5
        self.known_namedw = {k+1:v for k,v in list(self.known_w.items())}

    @ut.skipIf(GEOPANDAS_EXTINCT, 'Missing Geopandas')
    def test_linestrings(self):
        import geopandas 
        eberly = geopandas.read_file(pysal_examples.get_path("eberly_net.shp")).iloc[0:8]
        eberly_w = {0: [1,2,3],
                    1: [0,4],
                    2: [0,3,4,5],
                    3: [0,2,7],
                    4: [1,2,5],
                    5: [2,4,6],
                    6: [5],
                    7: [3]}
        eberly_w = W(neighbors=eberly_w).sparse.toarray()
        computed = self.cls.from_dataframe(eberly).sparse.toarray()
        np.testing.assert_array_equal(eberly_w, computed)

class Test_Rook(ut.TestCase, Contiguity_Mixin):
    def setUp(self):
        Contiguity_Mixin.setUp(self)
        
        self.known_w = {2: 1.0, 3: 1.0, 5: 1.0, 7: 1.0, 
                             8: 1.0, 10: 1.0, 14: 1.0}
        self.known_wi = 4
        self.cls = c.Rook
        self.idVariable = 'POLYID'
        self.known_name = 5
        self.known_namedw = {k+1:v for k,v in list(self.known_w.items())}

class Test_Voronoi(ut.TestCase):
    def test_voronoiW(self):
        np.random.seed(12345)
        points = np.random.random((5,2))*10 + 10
        w = c.Voronoi(points)
        self.assertEqual(w.n, 5)
        self.assertEqual(w.neighbors, {0: [2, 3, 4],
                                        1: [2], 2: [0, 1, 4],
                                        3: [0, 4], 4: [0, 2, 3]})

q = ut.TestLoader().loadTestsFromTestCase(Test_Queen)
r = ut.TestLoader().loadTestsFromTestCase(Test_Rook)
suite = ut.TestSuite([q, r])
if __name__ == '__main__':
    runner = ut.TextTestRunner()
    runner.run(suite)



"""
Base class for managing example datasets
"""

# Authors: Serge Rey
# License: BSD 3 Clause

import io
import os
import webbrowser
from os import environ, makedirs
from os.path import exists, expanduser, join
import zipfile
import requests
import pandas
from bs4 import BeautifulSoup
from ..io import open as ps_open

PYSALDATA = 'pysal_data'

def get_data_home():
    """Return the path of the libpysal data directory.


    This folder (~/pysal_data)is used by some large dataset loaders to avoid
    downloading the data multiple times.


    Alternatively, it can be set by the 'PYSALDATA' environment variable or
    programmatically by giving an explicit folder path. The '~' symbol is
    expanded to the user home folder

    If the folder does not already exisit, it is automatically created.
    """

    data_home = environ.get('PYSALDATA', join("~", PYSALDATA))
    data_home = expanduser(data_home)
    if not exists(data_home):
        makedirs(data_home)
    return data_home


def get_list_of_files(dir_name):
    """
    create a list of file and sub directories in dir_name
    """
    #names in the given directory
    file_list = os.listdir(dir_name)
    all_files = list()
    # Iterate over all the entries
    for entry in file_list:
        # Create full path
        full_path = os.path.join(dir_name, entry)
        # If entry is a directory then get the list of files in this directory
        if os.path.isdir(full_path):
            all_files = all_files + get_list_of_files(full_path)
        else:
            all_files.append(full_path)

    return all_files

def type_of_script():
    """Helper function to determine run context"""
    try:
        ipy_str = str(type(get_ipython()))
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        if 'terminal' in ipy_str:
            return 'ipython'
    except:
        return 'terminal'



class Example:
    """
    Example Dataset
    """
    def __init__(self, name, description, n, k, download_url,
                explain_url):
        self.name = name
        self.description = description
        self.n = n
        self.k = k
        self.download_url = download_url
        self.explain_url = explain_url
        self.root = name.replace(" ", "_")
        self.installed = self.downloaded()

    def get_local_path(self, path=get_data_home()):
        """Get local path for example"""
        return join(path, self.root)

    def get_path(self, file_name, verbose=True):
        """
        get path for local file
        """
        file_list = self.get_file_list()
        for file_path in file_list:
            base_name = os.path.basename(file_path)
            if file_name == base_name:
                return file_path
        if verbose:
            print("{} is not a file in this example".format(file_name))
        return None

    def downloaded(self):
        """
        Check if example has already been installed
        """
        path = self.get_local_path()
        if os.path.isdir(path):
            self.installed = True
            return True
        return False


    def explain(self):
        """
        Provide a description of the example.
        """
        file_name = self.explain_url.split("/")[-1]
        if file_name == 'README.md':
            explain_page = requests.get(self.explain_url)
            crawled = BeautifulSoup(explain_page.text, 'html.parser')
            print(crawled.text)
            return None
        if type_of_script() == 'terminal':
            webbrowser.open(self.explain_url)
            return None
        from IPython.display import IFrame
        return IFrame(self.explain_url, width=700, height=350)

    def download(self, path=get_data_home()):
        """
        Download the files for the example.
        """
        if self.downloaded():
            print('Already downloaded')
        else:
            request = requests.get(self.download_url)
            archive = zipfile.ZipFile(io.BytesIO(request.content))
            target = join(path, self.root)
            print('Downloading {} to {}'.format(self.name, target))
            archive.extractall(path=target)
            self.zipfile = archive
            self.installed = True

    def get_file_list(self):
        """
        Get list of local files for the example.
        """
        path = self.get_local_path()
        if os.path.isdir(path):
            return get_list_of_files(path)
        return None


    def json_dict(self):
        """
        container for example meta data
        """
        meta = {}
        meta['name'] = self.name
        meta['description'] = self.description
        meta['download_url'] = self.download_url
        meta['explain_url'] = self.explain_url
        meta['root'] = self.root
        return meta

    def load(self, file_name):
        """
        dispatch to libpysal.io to open file
        """
        pth = self.get_path(file_name)
        if pth:
            return ps_open(pth)




class Examples:
    """
    Manager for pysal example datasets.

    """

    def __init__(self):
        self.datasets = {}


    def add_examples(self, examples):
        """
        add examples to the set of datasets available
        """
        self.datasets.update(examples)


    def explain(self, example_name):
        if example_name in self.datasets:
            return self.datasets[example_name].explain()
        else:
            print('not available')

    def available(self):
        """
        report available datasets
        """
        datasets = self.datasets
        names = list(datasets.keys())
        names.sort()
        rows = []
        for name in names:
            description = datasets[name].description
            installed = datasets[name].installed
            rows.append([name, description, installed])
        datasets = pandas.DataFrame(data = rows, columns=['Name', 'Description', 'Installed'])
        datasets.style.set_properties(subset=['text'], **{'width': '300px'})
        print(datasets.to_string())


    def load(self, example_name):
        """
        load example dataset, download if not locally available
        """
        if example_name in self.datasets:
            example = self.datasets[example_name]
            if example.installed:
                return example
            else:
                "Downloading: {}".format(example_name)
                example.download()
                return example
        else:
            print('Example not available: {}'.format(example_name))
            return None

    def download_remotes(self):
        """
        Dowload all remotes

        """
        names = list(self.remotes.keys())
        names.sort()

        for name in names:
            print(name)
            example = self.remotes[name]
            try:
                example.download()
            except:
                print('Example not downloaded: {}'.format(name))


    def get_installed_names(self):
        """Return names of all currently installed datasets"""
        ds = self.datasets
        return [name for name in ds if ds[name].installed]

example_manager = Examples()




"""
The :mod:`libpysal.examples` module includes a number of small built-in example datasets as well as functions to fetch larger datasets.
"""
from .base import example_manager
from .remotes import datasets as remote_datasets
from .remotes import download as fetch_all
from .builtin import datasets as builtin_datasets

__all__ = ['get_path', 'available', 'explain', 'fetch_all']

example_manager.add_examples(remote_datasets)
example_manager.add_examples(builtin_datasets)


def available():
    """
    List available datasets
    """
    return example_manager.available()


def explain(name):
    """
    Explain a dataset by name
    """
    return example_manager.explain(name)


def load_example(example_name):
    """
    Load example dataset instance
    """
    return example_manager.load(example_name)


def get_path(file_name):
    """
    get the path for a file by searching installed datasets
    """

    installed = example_manager.get_installed_names()
    for name in installed:
        example = example_manager.datasets[name]
        pth = example.get_path(file_name, verbose=False)
        if pth:
            return pth
    print("{} is not a file in any installed dataset.".format(file_name))



import unittest
from ..weight_converter import WeightConverter
from ..weight_converter import weight_convert
from ...fileio import FileIO as psopen
import tempfile
import os
import warnings
#import pysal_examples
from .... import examples as pysal_examples

@unittest.skip('This function is deprecated')
class test_WeightConverter(unittest.TestCase):
    def setUp(self):
        self.base_dir = pysal_examples.get_path('')
        test_files = ['arcgis_ohio.dbf', 'arcgis_txt.txt', 'ohio.swm',
                           'wmat.dat', 'wmat.mtx', 'sids2.gal', 'juvenile.gwt',
                           'geobugs_scot', 'stata_full.txt', 'stata_sparse.txt',
                           'spat-sym-us.mat', 'spat-sym-us.wk1']
        self.test_files = [pysal_examples.get_path(f) for f in test_files]
        dataformats = ['arcgis_dbf', 'arcgis_text', None, None, None, None, None,
                       'geobugs_text', 'stata_text', 'stata_text', None, None]
        ns = [88, 3, 88, 49, 49, 100, 168, 56, 56, 56, 46, 46]
        self.dataformats = dict(list(zip(self.test_files, dataformats)))
        self.ns = dict(list(zip(self.test_files, ns)))
        self.fileformats = [('dbf', 'arcgis_dbf'), ('txt', 'arcgis_text'), ('swm', None),
                            ('dat', None), ('mtx', None), ('gal', None), ('',
                                                                          'geobugs_text'),
                            ('gwt', None), ('txt', 'stata_text'), ('mat', None), ('wk1', None)]

    def test__setW(self):
        for f in self.test_files:
            with warnings.catch_warnings(record=True) as warn:
                # note: we are just suppressing the warnings here; individual warnings
                #       are tested in their specific readers
                warnings.simplefilter("always")
                wc = WeightConverter(f,
                                     dataFormat=self.dataformats[f])
            self.assertEqual(wc.w_set(), True)
            self.assertEqual(wc.w.n, self.ns[f])

    def test_write(self):
        for f in self.test_files:
            with warnings.catch_warnings(record=True) as warn:
                # note: we are just suppressing the warnings here; individual warnings
                #       are tested in their specific readers
                warnings.simplefilter("always")
                wc = WeightConverter(f,
                                     dataFormat=self.dataformats[f])

            for ext, dataformat in self.fileformats:
                if f.lower().endswith(ext):
                    continue
                temp_f = tempfile.NamedTemporaryFile(
                    suffix='.%s' % ext, dir=self.base_dir)
                temp_fname = temp_f.name
                temp_f.close()

                with warnings.catch_warnings(record=True) as warn:
                    # note: we are just suppressing the warnings here; individual warnings
                    #       are tested in their specific readers
                    warnings.simplefilter("always")
                    if ext == 'swm':
                        wc.write(temp_fname, useIdIndex=True)
                    elif dataformat is None:
                        wc.write(temp_fname)
                    elif dataformat in ['arcgis_dbf', 'arcgis_text']:
                        wc.write(temp_fname, dataFormat=dataformat,
                                 useIdIndex=True)
                    elif dataformat == 'stata_text':
                        wc.write(temp_fname, dataFormat=dataformat,
                                 matrix_form=True)
                    else:
                        wc.write(temp_fname, dataFormat=dataformat)

                with warnings.catch_warnings(record=True) as warn:
                    # note: we are just suppressing the warnings here; individual warnings
                    #       are tested in their specific readers
                    warnings.simplefilter("always")
                    if dataformat is None:
                        wnew = psopen(temp_fname, 'r').read()
                    else:
                        wnew = psopen(temp_fname, 'r', dataformat).read()

                if (ext in ['dbf', 'swm', 'dat', 'wk1', 'gwt'] or dataformat == 'arcgis_text'):
                    self.assertEqual(wnew.n, wc.w.n - len(wc.w.islands))
                else:
                    self.assertEqual(wnew.n, wc.w.n)
                os.remove(temp_fname)

    def test_weight_convert(self):
        for f in self.test_files:
            inFile =  f
            inDataFormat = self.dataformats[f]
            with warnings.catch_warnings(record=True) as warn:
                # note: we are just suppressing the warnings here; individual warnings
                #       are tested in their specific readers
                warnings.simplefilter("always")
                if inDataFormat is None:
                    in_file = psopen(inFile, 'r')
                else:
                    in_file = psopen(inFile, 'r', inDataFormat)
                wold = in_file.read()
                in_file.close()

            for ext, dataformat in self.fileformats:
                if f.lower().endswith(ext):
                    continue
                temp_f = tempfile.NamedTemporaryFile(
                    suffix='.%s' % ext, dir=self.base_dir)
                outFile = temp_f.name
                temp_f.close()
                outDataFormat, useIdIndex, matrix_form = dataformat, False, False
                if ext == 'swm' or dataformat in ['arcgis_dbf', 'arcgis_text']:
                    useIdIndex = True
                elif dataformat == 'stata_text':
                    matrix_form = True

                with warnings.catch_warnings(record=True) as warn:
                    # note: we are just suppressing the warnings here; individual warnings
                    #       are tested in their specific readers
                    warnings.simplefilter("always")
                    weight_convert(inFile, outFile, inDataFormat, outDataFormat, useIdIndex, matrix_form)

                with warnings.catch_warnings(record=True) as warn:
                    # note: we are just suppressing the warnings here; individual warnings
                    #       are tested in their specific readers
                    warnings.simplefilter("always")
                    if dataformat is None:
                        wnew = psopen(outFile, 'r').read()
                    else:
                        wnew = psopen(outFile, 'r', dataformat).read()

                if (ext in ['dbf', 'swm', 'dat', 'wk1', 'gwt'] or dataformat == 'arcgis_text'):
                    self.assertEqual(wnew.n, wold.n - len(wold.islands))
                else:
                    self.assertEqual(wnew.n, wold.n)
                os.remove(outFile)

if __name__ == '__main__':
    unittest.main()



from ...common import requires
from ...cg import asShape
from .file import read_files, write_files
import shapely.geometry as sgeom
import pandas as pd

@requires('geopandas')
def geopandas(filename, **kw):
    import geopandas
    return geopandas.read_file(filename, **kw)

@requires('fiona')
def fiona(filename, geom_type='shapely', **kw):
    if geom_type == 'shapely':
        converter = sgeom.shape
    elif geom_type is None:
        converter = lambda x: x
    else:
        converter = asShape
    import fiona
    props = {}
    with fiona.open(filename, **kw) as f:
        for i,feat in enumerate(f):
            idx = feat.get('id', i)
            try:
                idx = int(idx)
            except ValueError:
                pass
            props.update({idx:feat.get('properties', dict())})
            props[idx].update({'geometry':converter(feat['geometry'])})
    return pd.DataFrame().from_dict(props).T

_readers = {'read_shapefile':read_files, 
            'read_fiona':fiona}
_writers = {'to_shapefile':write_files}

_pandas_readers = {k:v for k,v in list(pd.io.api.__dict__.items()) if k.startswith('read_')}



"""Unit test for util.py"""
from ..util import lat2W, nonplanar_neighbors, fuzzy_contiguity
from .. import util
from ..weights import W, WSP
from ..distance import DistanceBand, KNN
from ..contiguity import Queen, Rook
from ...io.fileio import FileIO as psopen
from ... import examples
import numpy as np
import unittest


try:
    import geopandas as gpd
    HAS_GEOPANDAS = True
except:
    HAS_GEOPANDAS = False


class Testutil(unittest.TestCase):
    def setUp(self):
        self.w = Rook.from_shapefile(
            examples.get_path('10740.shp'))

        self.rio = examples.load_example('Rio Grande do Sul')

    def test_lat2W(self):
        w9 = lat2W(3, 3)
        self.assertEqual(w9.pct_nonzero, 29.62962962962963)
        self.assertEqual(w9[0], {1: 1.0, 3: 1.0})
        self.assertEqual(w9[3], {0: 1.0, 4: 1.0, 6: 1.0})

    def test_lat2SW(self):
        w9 = util.lat2SW(3, 3)
        rows, cols = w9.shape
        n = rows * cols
        pct_nonzero = w9.nnz / float(n)
        self.assertEqual(pct_nonzero, 0.29629629629629628)
        data = w9.todense().tolist()
        self.assertEqual(data[0], [0, 1, 0, 1, 0, 0, 0, 0, 0])
        self.assertEqual(data[1], [1, 0, 1, 0, 1, 0, 0, 0, 0])
        self.assertEqual(data[2], [0, 1, 0, 0, 0, 1, 0, 0, 0])
        self.assertEqual(data[3], [1, 0, 0, 0, 1, 0, 1, 0, 0])
        self.assertEqual(data[4], [0, 1, 0, 1, 0, 1, 0, 1, 0])
        self.assertEqual(data[5], [0, 0, 1, 0, 1, 0, 0, 0, 1])
        self.assertEqual(data[6], [0, 0, 0, 1, 0, 0, 0, 1, 0])
        self.assertEqual(data[7], [0, 0, 0, 0, 1, 0, 1, 0, 1])
        self.assertEqual(data[8], [0, 0, 0, 0, 0, 1, 0, 1, 0])

    def test_block_weights(self):
        regimes = np.ones(25)
        regimes[list(range(10, 20))] = 2
        regimes[list(range(21, 25))] = 3
        regimes = np.array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
                            2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 1., 3., 3.,
                            3., 3.])
        w = util.block_weights(regimes)
        ww0 = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        self.assertEqual(w.weights[0], ww0)
        wn0 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 20]
        self.assertEqual(w.neighbors[0], wn0)
        regimes = ['n', 'n', 's', 's', 'e', 'e', 'w', 'w', 'e']
        w = util.block_weights(regimes)
        wn = {0: [1], 1: [0], 2: [3], 3: [2], 4: [5, 8], 5: [4, 8],
              6: [7], 7: [6], 8: [4, 5]}
        self.assertEqual(w.neighbors, wn)
        ids = ['id-%i'%i for i in range(len(regimes))]
        w = util.block_weights(regimes, ids=np.array(ids))
        w0 = {'id-1': 1.0}
        self.assertEqual(w['id-0'], w0)
        w = util.block_weights(regimes, ids=ids)
        w0 = {'id-1': 1.0}
        self.assertEqual(w['id-0'], w0)

    def test_comb(self):
        x = list(range(4))
        l = []
        for i in util.comb(x, 2):
            l.append(i)
        lo = [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]
        self.assertEqual(l, lo)

    def test_order(self):
        w3 = util.order(self.w, kmax=3)
        w3105 = [1, -1, 1, 2, 1]
        self.assertEqual(w3105, w3[1][0:5])

    def test_higher_order(self):
        w10 = lat2W(10, 10)
        w10_2 = util.higher_order(w10, 2)
        w10_20 = {2: 1.0, 11: 1.0, 20: 1.0}
        self.assertEqual(w10_20, w10_2[0])
        w5 = lat2W()
        w50 = {1: 1.0, 5: 1.0}
        self.assertEqual(w50, w5[0])
        w51 = {0: 1.0, 2: 1.0, 6: 1.0}
        self.assertEqual(w51, w5[1])
        w5_2 = util.higher_order(w5, 2)
        w5_20 = {2: 1.0, 10: 1.0, 6: 1.0}
        self.assertEqual(w5_20, w5_2[0])

    def test_higher_order_classes(self):
        wdb = DistanceBand.from_shapefile(examples.get_path('baltim.shp'), 34)
        wknn = KNN.from_shapefile(examples.get_path('baltim.shp'), 10)
        wrook = Rook.from_shapefile(examples.get_path('columbus.shp'))
        wqueen = Queen.from_shapefile(examples.get_path('columbus.shp'))
        wsparse = wqueen.sparse
        ww = W(wknn.neighbors, wknn.weights)
        util.higher_order(wdb, 2)
        util.higher_order(wknn, 3)
        util.higher_order(wrook, 4)
        util.higher_order(wqueen, 5)
        util.higher_order(wsparse, 2)
        util.higher_order(ww, 2)
        ww.transform = 'r'
        wsparse_notbinary = wrook.sparse
        with self.assertRaises(ValueError):
            util.higher_order(wsparse, 2)
            util.higher_order(ww, 3)



    def test_shimbel(self):
        w5 = lat2W()
        w5_shimbel = util.shimbel(w5)
        w5_shimbel024 = 8
        self.assertEqual(w5_shimbel024, w5_shimbel[0][24])
        w5_shimbel004 = [-1, 1, 2, 3]
        self.assertEqual(w5_shimbel004, w5_shimbel[0][0:4])

    def test_full(self):
        neighbors = {'first': ['second'], 'second': ['first',
                                                     'third'], 'third': ['second']}
        weights = {'first': [1], 'second': [1, 1], 'third': [1]}
        w = W(neighbors, weights)
        wf, ids = util.full(w)
        wfo = np.array([[0., 1., 0.], [1., 0., 1.], [0., 1., 0.]])
        np.testing.assert_array_almost_equal(wfo, wf, decimal=8)
        idso = ['first', 'second', 'third']
        self.assertEqual(idso, ids)

    def test_full2W(self):
        a = np.zeros((4, 4))
        for i in range(len(a)):
            for j in range(len(a[i])):
                if i != j:
                    a[i, j] = np.random.random(1)
        w = util.full2W(a)
        np.testing.assert_array_equal(w.full()[0], a)
        ids = ['myID0', 'myID1', 'myID2', 'myID3']
        w = util.full2W(a, ids=ids)
        np.testing.assert_array_equal(w.full()[0], a)
        w.full()[0] == a

    def test_WSP2W(self):
        sp = util.lat2SW(2, 5)
        wsp = WSP(sp)
        w = util.WSP2W(wsp)
        self.assertEqual(w.n, 10)
        self.assertEqual(w[0], {1: 1, 5: 1})
        w = psopen(examples.get_path('sids2.gal'), 'r').read()
        wsp = WSP(w.sparse, w.id_order)
        w = util.WSP2W(wsp)
        self.assertEqual(w.n, 100)
        self.assertEqual(w['37135'], {'37001': 1.0, '37033': 1.0,
                                       '37037': 1.0, '37063': 1.0, '37145': 1.0})

    def test_insert_diagonal(self):
        w1 = util.insert_diagonal(self.w)
        r1 = {0: 1.0, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        self.assertEqual(w1[0], r1)
        w1 = util.insert_diagonal(self.w, 20)
        r1 = {0: 20, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        self.assertEqual(w1[0], r1)
        diag = np.arange(100, 100 + self.w.n)
        w1 = util.insert_diagonal(self.w, diag)
        r1 = {0: 100, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        self.assertEqual(w1[0], r1)

    def test_remap_ids(self):
        w = lat2W(3, 2)
        wid_order = [0, 1, 2, 3, 4, 5]
        self.assertEqual(wid_order, w.id_order)
        wneighbors0 = [2, 1]
        self.assertEqual(wneighbors0, w.neighbors[0])
        old_to_new = {0: 'a', 1: 'b', 2: 'c', 3: 'd', 4: 'e', 5: 'f'}
        w_new = util.remap_ids(w, old_to_new)
        w_newid_order = ['a', 'b', 'c', 'd', 'e', 'f']
        self.assertEqual(w_newid_order, w_new.id_order)
        w_newdneighborsa = ['c', 'b']
        self.assertEqual(w_newdneighborsa, w_new.neighbors['a'])

    def test_get_ids_shp(self):
        polyids = util.get_ids(
            examples.get_path('columbus.shp'), "POLYID")
        polyids5 = [1, 2, 3, 4, 5]
        self.assertEqual(polyids5, polyids[:5])

    @unittest.skipIf(not HAS_GEOPANDAS, "Missing geopandas, cannot test get_ids with gdf")
    def test_get_ids_gdf(self):
        gdf = gpd.read_file(examples.get_path('columbus.shp'))
        polyids = util.get_ids(gdf, "POLYID")
        polyids5 = [1, 2, 3, 4, 5]
        self.assertEqual(polyids5, polyids[:5])

    def test_get_points_array_from_shapefile(self):
        xy = util.get_points_array_from_shapefile(
            examples.get_path('juvenile.shp'))
        xy3 = np.array([[94., 93.], [80., 95.], [79., 90.]])
        np.testing.assert_array_almost_equal(xy3, xy[:3], decimal=8)
        xy = util.get_points_array_from_shapefile(
            examples.get_path('columbus.shp'))
        xy3 = np.array([[8.82721847, 14.36907602], [8.33265837,
                                                    14.03162401], [9.01226541, 13.81971908]])
        np.testing.assert_array_almost_equal(xy3, xy[:3], decimal=8)

    def test_min_threshold_distance(self):
        x, y = np.indices((5, 5))
        x.shape = (25, 1)
        y.shape = (25, 1)
        data = np.hstack([x, y])
        mint = 1.0
        self.assertEqual(
            mint, util.min_threshold_distance(data))

    def test_attach_islands(self):
        w = Rook.from_shapefile(examples.get_path('10740.shp'))
        w_knn1 = KNN.from_shapefile(examples.get_path('10740.shp'), k=1)
        w_attach = util.attach_islands(w, w_knn1)
        self.assertEqual(w_attach.islands, [])
        self.assertEqual(w_attach[w.islands[0]], {166: 1.0})

    @unittest.skipIf(not HAS_GEOPANDAS, "Missing geopandas, cannot test nonplanar neighbors")
    def test_nonplanar_neighbors(self):
        df = gpd.read_file(examples.get_path('map_RS_BR.shp'))
        w = Queen.from_dataframe(df)
        self.assertEqual(w.islands, [0, 4, 23, 27, 80, 94, 101, 107, 109, 119, 122, 139, 169, 175, 223, 239, 247, 253, 254, 255, 256, 261, 276, 291, 294, 303, 321, 357, 374])
        wnp = nonplanar_neighbors(w, df)
        self.assertEqual(wnp.islands, [])
        self.assertEqual(w.neighbors[0], [])
        self.assertEqual(wnp.neighbors[0], [23, 59, 152, 239])
        self.assertEqual(wnp.neighbors[23], [0, 45, 59, 107, 152, 185, 246])

    @unittest.skipIf(not HAS_GEOPANDAS, "Missing geopandas, cannot test fuzzy_contiguity")
    def test_fuzzy_contiguity(self):
        rs = examples.get_path('map_RS_BR.shp')
        rs_df = gpd.read_file(rs)
        wf = fuzzy_contiguity(rs_df)
        self.assertEqual(wf.islands, [])
        self.assertEqual(set(wf.neighbors[0]), set([239, 59, 152, 23, 107]))


suite = unittest.TestLoader().loadTestsFromTestCase(Testutil)

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite)



"""
Computation of alpha shape algorithm in 2-D based on original implementation
by Tim Kittel (@timkittel) available at:

    https://github.com/timkittel/alpha-shapes

Author(s):
    Dani Arribas-Bel daniel.arribas.bel@gmail.com
"""

import numpy as np
import scipy.spatial as spat

from ..common import requires, jit, HAS_JIT

if not HAS_JIT:
    from warnings import warn
    NUMBA_WARN = "Numba not imported, so alpha shape construction may be slower than expected."

EPS = np.finfo(float).eps

__all__ = ['alpha_shape', 'alpha_shape_auto']

@jit
def nb_dist(x, y):
    '''
    numba implementation of distance between points `x` and `y`
    ...

    Parameters
    ----------
    
    x       : ndarray
              Coordinates of point `x`
    y       : ndarray
              Coordinates of point `y`

    Returns
    -------
    
    dist    : float
              Distance between `x` and `y`

    Examples
    --------

    >>> x = np.array([0, 0])
    >>> y = np.array([1, 1])
    >>> dist = nb_dist(x, y)
    >>> dist
    1.4142135623730951
    
    '''
    sum = 0
    for x_i, y_i in zip(x, y):
        sum += (x_i - y_i)**2
    dist = np.sqrt(sum)
    return dist

@jit(nopython=True)
def r_circumcircle_triangle_single(a, b, c):
    '''
    Computation of the circumcircle of a single triangle
    ...

    Source for equations:

    > https://www.mathopenref.com/trianglecircumcircle.html

    [Last accessed July 11th. 2018]

    Parameters
    ----------
    
    a       : ndarray
              (2,) Array with coordinates of vertex `a` of the triangle
    b       : ndarray
              (2,) Array with coordinates of vertex `b` of the triangle
    c       : ndarray
              (2,) Array with coordinates of vertex `c` of the triangle

    Returns
    -------
    
    r       : float
              Circumcircle of the triangle

    Examples
    --------

    >>> a = np.array([0, 0])
    >>> b = np.array([0.5, 0])
    >>> c = np.array([0.25, 0.25])
    >>> r = r_circumcircle_triangle_single(a, b, c)
    >>> r
    0.2500000000000001
    
    '''
    ab = nb_dist(a, b)
    bc = nb_dist(b, c)
    ca = nb_dist(c, a)

    num = ab * bc * ca
    den = np.sqrt( (ab + bc + ca) * \
                   (bc + ca - ab) * \
                   (ca + ab - bc) * \
                   (ab + bc - ca) )
    if den == 0:
        return np.array([ab, bc, ca]).max() / 2.0
    else:
        return num / den

@jit(nopython=True)
def r_circumcircle_triangle(a_s, b_s, c_s):
    '''
    Computation of circumcircles for a series of triangles
    ...

    Parameters
    ----------
    
    a_s     : ndarray
              (N, 2) array with coordinates of vertices `a` of the triangles
    b_s     : ndarray
              (N, 2) array with coordinates of vertices `b` of the triangles
    c_s     : ndarray
              (N, 2) array with coordinates of vertices `c` of the triangles

    Returns
    -------
    
    radii   : ndarray
              (N,) array with circumcircles for every triangle

    Examples
    --------

    >>> a_s = np.array([[0, 0], [2, 1], [3, 2]])
    >>> b_s = np.array([[1, 0], [5, 1], [2, 4]])
    >>> c_s = np.array([[0, 7], [1, 3], [4, 2]])
    >>> rs = r_circumcircle_triangle(a_s, b_s, c_s)
    >>> rs
    array([3.53553391, 2.5       , 1.58113883])
    '''
    len_a = len(a_s)
    r2 = np.zeros( (len_a,) )
    for i in range(len_a):
        r2[i] = r_circumcircle_triangle_single(a_s[i],
                                               b_s[i],
                                               c_s[i])
    return r2

@jit
def get_faces(triangle):
    '''
    Extract faces from a single triangle
    ...

    Parameters
    ----------
    
    triangles       : ndarray
                      (3,) array with the vertex indices for a triangle

    Returns
    -------
    
    faces           : ndarray
                      (3, 2) array with a row for each face containing the
                      indices of the two points that make up the face

    Examples
    --------
    
    >>> triangle = np.array([3, 1, 4], dtype=np.int32)
    >>> faces = get_faces(triangle)
    >>> faces
    array([[3., 1.],
           [1., 4.],
           [4., 3.]])

    '''
    faces = np.zeros((3, 2))
    for i, (i0, i1) in enumerate([(0, 1), (1, 2), (2, 0)]):
        faces[i] = triangle[i0], triangle[i1]
    return faces

@jit
def build_faces(faces, triangles_is,
        num_triangles, num_faces_single):
    '''
    Build facing triangles

    ...

    Parameters
    ----------
    
    faces               : ndarray
                          (num_triangles * num_faces_single, 2) array of
                          zeroes in int form
    triangles_is        : ndarray
                          (D, 3) array, where D is the number of Delaunay
                          triangles, with the vertex indices for each
                          triangle
    num_triangles       : int
                          Number of triangles
    num_faces_single    : int
                          Number of faces a triangle has (i.e. 3)

    Returns
    -------
    
    faces               : ndarray
                          Two dimensional array with a row for every facing
                          segment containing the indices of the coordinate points

    Examples
    --------
    
    >>> import scipy.spatial as spat
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> triangulation = spat.Delaunay(pts)
    >>> triangulation.simplices
    array([[3, 1, 4],
           [1, 2, 4],
           [2, 1, 0]], dtype=int32)
    >>> num_faces_single = 3
    >>> num_triangles = triangulation.simplices.shape[0]
    >>> num_faces = num_triangles * num_faces_single
    >>> faces = np.zeros((num_faces, 2), dtype=np.int_)
    >>> mask = np.ones((num_faces,), dtype=np.bool_)
    >>> faces = build_faces(faces, triangulation.simplices, num_triangles, num_faces_single)
    >>> faces
    array([[3, 1],
           [1, 4],
           [4, 3],
           [1, 2],
           [2, 4],
           [4, 1],
           [2, 1],
           [1, 0],
           [0, 2]])

    '''
    for i in range(num_triangles):
        from_i = num_faces_single * i
        to_i = num_faces_single * (i+1)
        faces[from_i: to_i] = get_faces(triangles_is[i])
    return faces

@jit
def nb_mask_faces(mask, faces):
    '''
    Run over each row in `faces`, if the face in the following row is the
    same, then mark both as False on `mask`
    ...

    Parameters
    ----------
    
    mask    : ndarray
              One-dimensional boolean array set to True with as many
              observations as rows in `faces`
    faces   : ndarray
              Sorted sequence of faces for all triangles (ie. triangles split
              by each segment)

    Returns
    -------
    
    masked  : ndarray
              Sequence of outward-facing faces

    Examples
    --------
    
    >>> import numpy as np
    >>> faces = np.array([[0, 1], [0, 2], [1, 2], [1, 2], [1, 3], [1, 4], [1, 4], [2, 4], [3, 4]])
    >>> mask = np.ones((faces.shape[0], ), dtype=np.bool_)
    >>> masked = nb_mask_faces(mask, faces)
    >>> masked
    array([[0, 1],
           [0, 2],
           [1, 3],
           [2, 4],
           [3, 4]])
    
    '''
    for k in range(faces.shape[0]-1):
        if mask[k]:
            if np.all(faces[k] == faces[k+1]):
                mask[k] = False
                mask[k+1] = False
    return faces[mask]

def get_single_faces(triangles_is):
    '''
    Extract outward facing edges from collection of triangles
    ...

    Parameters
    ----------
    
    triangles_is    : ndarray
                      (D, 3) array, where D is the number of Delaunay triangles,
                      with the vertex indices for each triangle

    Returns
    -------
    
    single_faces    : ndarray

    Example
    -------
    
    >>> import scipy.spatial as spat
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> alpha = 0.33
    >>> triangulation = spat.Delaunay(pts)
    >>> triangulation.simplices
    array([[3, 1, 4],
           [1, 2, 4],
           [2, 1, 0]], dtype=int32)
    >>> get_single_faces(triangulation.simplices)
    array([[0, 1],
           [0, 2],
           [1, 3],
           [2, 4],
           [3, 4]])

    '''
    num_faces_single = 3
    num_triangles = triangles_is.shape[0]
    num_faces = num_triangles * num_faces_single
    faces = np.zeros((num_faces, 2), dtype=np.int_)
    mask = np.ones((num_faces,), dtype=np.bool_)

    faces = build_faces(faces, triangles_is,
                        num_triangles, num_faces_single)

    orderlist = ["x{}".format(i) for i in range(faces.shape[1])]
    dtype_list = [(el, faces.dtype.str) for el in orderlist]
    # Arranging each face so smallest vertex is first
    faces.sort(axis=1)
    # Arranging faces in ascending way
    faces.view(dtype_list).sort(axis=0)
    # Masking
    single_faces = nb_mask_faces(mask, faces)
    return single_faces

@requires('geopandas', 'shapely')
def alpha_geoms(alpha, triangles, radii, xys):
    '''
    Generate alpha-shape polygon(s) from `alpha` value, vertices of `triangles`,
    the `radii` for all points, and the points themselves
    ...

    Parameters
    ----------
    
    alpha       : float
                  Alpha value to delineate the alpha-shape
    triangles   : ndarray
                  (D, 3) array, where D is the number of Delaunay triangles,
                  with the vertex indices for each triangle
    radii       : ndarray
                  (N,) array with circumcircles for every triangle
    xys         : ndarray
                  (N, 2) array with one point per row and coordinates structured
                  as X and Y

    Returns
    -------
    
    geoms       : GeoSeries
                  Polygon(s) resulting from the alpha shape algorithm. The
                  GeoSeries object remains so even if only a single polygon is
                  returned. There is no CRS included in the object.

    Examples
    --------
    
    >>> import scipy.spatial as spat
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> alpha = 0.33
    >>> triangulation = spat.Delaunay(pts)
    >>> triangles = pts[triangulation.simplices]
    >>> triangles
    array([[[6, 7],
            [3, 5],
            [9, 3]],
    <BLANKLINE>
           [[3, 5],
            [4, 1],
            [9, 3]],
    <BLANKLINE>
           [[4, 1],
            [3, 5],
            [0, 1]]])
    >>> a_pts = triangles[:, 0, :]
    >>> b_pts = triangles[:, 1, :]
    >>> c_pts = triangles[:, 2, :]
    >>> radii = r_circumcircle_triangle(a_pts, b_pts, c_pts)
    >>> geoms = alpha_geoms(alpha, triangulation.simplices, radii, pts)
    >>> geoms
    0    POLYGON ((0.00000 1.00000, 3.00000 5.00000, 4....
    dtype: geometry
    
    '''
    from shapely.geometry import LineString
    from shapely.ops import polygonize
    from geopandas import GeoSeries

    triangles_reduced = triangles[radii < 1/alpha]
    outer_triangulation = get_single_faces(triangles_reduced)
    face_pts = xys[outer_triangulation]
    geoms = GeoSeries(list(polygonize(list(map(LineString,
                                               face_pts)))))
    return geoms

@requires('geopandas', 'shapely')
def alpha_shape(xys, alpha):
    '''
    Alpha-shape delineation (Edelsbrunner, Kirkpatrick &
    Seidel, 1983) from a collection of points
    ...

    Parameters
    ----------
    
    xys     : ndarray
              (N, 2) array with one point per row and coordinates structured as X
              and Y
    alpha   : float
              Alpha value to delineate the alpha-shape

    Returns
    -------
    
    shapes  : GeoSeries
              Polygon(s) resulting from the alpha shape algorithm. The
              GeoSeries object remains so even if only a single polygon is
              returned. There is no CRS included in the object.

    Examples
    --------

    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> alpha = 0.1
    >>> poly = alpha_shape(pts, alpha)
    >>> poly
    0    POLYGON ((0.00000 1.00000, 3.00000 5.00000, 6....
    dtype: geometry
    >>> poly.centroid
    0    POINT (4.69048 3.45238)
    dtype: geometry


    References
    ----------

    Edelsbrunner, H., Kirkpatrick, D., & Seidel, R. (1983). On the shape of
        a set of points in the plane. IEEE Transactions on information theory,
        29(4), 551-559.
    
    '''
    if not HAS_JIT:
        warn(NUMBA_WARN)
    if xys.shape[0] < 4:
        from shapely import ops, geometry as geom
        return ops.cascaded_union([geom.Point(xy)
                                   for xy in xys])\
                  .convex_hull.buffer(0)
    triangulation = spat.Delaunay(xys)
    triangles = xys[triangulation.simplices]
    a_pts = triangles[:, 0, :]
    b_pts = triangles[:, 1, :]
    c_pts = triangles[:, 2, :]
    radii = r_circumcircle_triangle(a_pts, b_pts, c_pts)
    del triangles, a_pts, b_pts, c_pts
    geoms = alpha_geoms(alpha, triangulation.simplices, radii, xys)
    return geoms

def _valid_hull(geoms, points):
    '''
    Sanity check within ``alpha_shape_auto()`` to verify the generated
    alpha shape actually contains the original set of points (xys).
    
    Parameters
    ----------
    
    geoms   : GeoSeries
              see alpha_geoms()
    points  : list
              xys parameter cast as shapely.geometry.Point objects
    
    Returns
    -------
    
    flag    : bool
              Valid hull for alpha shape [True] or not [False]
    
    '''
    flag = True
    # if there is not exactly one polygon
    if geoms.shape[0] != 1:
        flag = False
    # if any (xys) points do not intersect the polygon
    for point in points:
        if not point.intersects(geoms[0]):
            flag = False
    return flag

@requires('geopandas', 'shapely')
def alpha_shape_auto(xys, step=1, verbose=False):
    '''
    Computation of alpha-shape delineation with automated selection of alpha.
    ...

    This method uses the algorithm proposed by  Edelsbrunner, Kirkpatrick &
    Seidel (1983) to return the tightest polygon that contains all points in
    `xys`. The algorithm ranks every point based on its radious and iterates
    over each point, checking whether the maximum alpha that would keep the
    point and all the other ones in the set with smaller radii results in a
    single polygon. If that is the case, it moves to the next point;
    otherwise, it retains the previous alpha value and returns the polygon
    as `shapely` geometry.

    Parameters
    ----------
    
    xys     : ndarray
              Nx2 array with one point per row and coordinates structured as X
              and Y
    step    : int
              [Optional. Default=1]
              Number of points in `xys` to jump ahead after checking whether the
              largest possible alpha that includes the point and all the
              other ones with smaller radii
    verbose : Boolean
              [Optional. Default=False] If True, it prints alpha values being
              tried at every step.


    Returns
    -------
    poly    : shapely.Polygon
              Tightest alpha-shape polygon containing all points in `xys`

    Examples
    --------

    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> poly = alpha_shape_auto(pts)
    >>> poly.bounds
    (0.0, 1.0, 9.0, 7.0)
    >>> poly.centroid.x, poly.centroid.y
    (4.690476190476191, 3.4523809523809526)

    References
    ----------

    Edelsbrunner, H., Kirkpatrick, D., & Seidel, R. (1983). On the shape of
        a set of points in the plane. IEEE Transactions on information theory,
        29(4), 551-559.
    
    '''
    if not HAS_JIT:
        warn(NUMBA_WARN)
    from shapely import geometry as geom
    if xys.shape[0] < 4:
        from shapely import ops
        return ops.cascaded_union([geom.Point(xy)
                                   for xy in xys])\
                  .convex_hull.buffer(0)
    triangulation = spat.Delaunay(xys)
    triangles = xys[triangulation.simplices]
    a_pts = triangles[:, 0, :]
    b_pts = triangles[:, 1, :]
    c_pts = triangles[:, 2, :]
    radii = r_circumcircle_triangle(a_pts, b_pts, c_pts)
    radii[np.isnan(radii)] = 0 # "Line" triangles to be kept for sure
    del triangles, a_pts, b_pts, c_pts
    radii_sorted_i = radii.argsort()
    triangles = triangulation.simplices[radii_sorted_i][::-1]
    radii = radii[radii_sorted_i][::-1]
    geoms_prev = alpha_geoms((1/radii.max())-EPS, triangles, radii, xys)
    xys_bb = np.array([*xys.min(axis=0), *xys.max(axis=0)])
    points = [geom.Point(pnt) for pnt in xys]
    if verbose:
        print('Step set to %i'%step)
    for i in range(0, len(radii), step):
        radi = radii[i]
        alpha = (1 / radi) - EPS
        if verbose:
            print('%.2f%% | Trying a = %f'\
            %((i+1)/radii.shape[0], alpha))
        geoms = alpha_geoms(alpha, triangles, radii, xys)
        if _valid_hull(geoms, points):
            geoms_prev = geoms
        else:
            break
    if verbose:
        print(geoms_prev.shape)
    return geoms_prev[0] # Return a shapely polygon

if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import time
    import geopandas as gpd
    plt.close('all')
    xys = np.random.random((1000, 2))
    t0 = time.time()
    geoms = alpha_shape_auto(xys, 1)
    t1 = time.time()
    print('%.2f Seconds to run algorithm'%(t1-t0))
    f, ax = plt.subplots(1)
    gpd.GeoDataFrame({'geometry':[geoms]}).plot(ax=ax, color='orange', alpha=0.5)
    ax.scatter(xys[:, 0], xys[:, 1], s=0.1)
    plt.show()




from unittest import TestCase, skipIf
from ...examples import get_path
from ..alpha_shapes import alpha_shape, alpha_shape_auto
import numpy as np
import os

try:
    import geopandas
    from shapely import geometry
    GEOPANDAS_EXTINCT = False
except ImportError:
    GEOPANDAS_EXTINCT = True

this_directory = os.path.dirname(__file__)


@skipIf(GEOPANDAS_EXTINCT, 'Geopandas is missing, so test will not run')
class Test_Alpha_Shapes(TestCase):
    def setUp(self):
        eberly = geopandas.read_file(get_path('eberly_net.shp'))
        eberly_vertices = eberly.geometry.apply(lambda x: np.hstack(x.xy).reshape(2, 2).T).values
        eberly_vertices = np.vstack(eberly_vertices)
        self.vertices = eberly_vertices

        self.a05 = geopandas.read_file(os.path.join(this_directory, 'data/alpha_05.shp')).geometry.to_numpy().item()
        self.a10 = geopandas.read_file(os.path.join(this_directory, 'data/alpha_tenth.shp')).geometry.to_numpy().item()
        self.a2  = geopandas.read_file(os.path.join(this_directory, 'data/alpha_fifth.shp')).geometry.to_numpy().item()
        self.a25 = geopandas.read_file(os.path.join(this_directory, 'data/alpha_fourth.shp')).geometry.to_numpy().item()
        self.a25 = geopandas.read_file(os.path.join(this_directory, 'data/alpha_fourth.shp')).geometry.to_numpy().item()

        self.autoalpha = geopandas.read_file(os.path.join(this_directory, 'data/alpha_auto.shp')).geometry[0]
    def test_alpha_shapes(self):
        new_a05 = alpha_shape(self.vertices, .05).to_numpy().item()
        new_a10 = alpha_shape(self.vertices, .10).to_numpy().item()
        new_a2  = alpha_shape(self.vertices, .2).to_numpy().item()
        new_a25 = alpha_shape(self.vertices, .25).to_numpy().item()

        assert new_a05.equals(self.a05)
        assert new_a10.equals(self.a10)
        assert new_a2.equals(self.a2)
        assert new_a25.equals(self.a25)

    def test_auto(self):
        auto_alpha = alpha_shape_auto(self.vertices, 5)

        assert self.autoalpha.equals(auto_alpha)

    def test_small_n(self):
        new_singleton = alpha_shape(self.vertices[0].reshape(1,-1), .5)
        assert isinstance(new_singleton, geometry.Polygon)
        new_duo = alpha_shape(self.vertices[:1], .5)
        assert isinstance(new_duo, geometry.Polygon)
        new_triple = alpha_shape(self.vertices[:2], .5)
        assert isinstance(new_triple, geometry.Polygon)



from .. import utils
from ..file import read_files as rf
import unittest as ut

@ut.skip('skpping converters and metadata inserters')
class Test_Utils(ut.TestCase):
    
    def test_converters(self):
        ## make a round trip to geodataframe and back
        raise Exception
    def test_insert_metadata(self):
        ## add an attribute to a dataframe and see 
        ## if it is pervasive over copies
        raise Exception



"""
Weights.
"""
__author__ = "Sergio J. Rey <srey@asu.edu> "

import copy
from os.path import basename as BASENAME
import math
import warnings
import numpy as np
import scipy.sparse
from scipy.sparse.csgraph import connected_components
#from .util import full, WSP2W resolve import cycle by
#forcing these into methods
from . import adjtools
from ..io.fileio import FileIO as popen

__all__ = ['W', 'WSP']


class W(object):
    """
    Spatial weights class.

    Parameters
    ----------
    neighbors            : dictionary
                           Key is region ID, value is a list of neighbor IDS.
                           Example:  {'a':['b'],'b':['a','c'],'c':['b']}
    weights              : dictionary
                           Key is region ID, value is a list of edge weights.
                           If not supplied all edge weights are assumed to have a weight of 1.
                           Example: {'a':[0.5],'b':[0.5,1.5],'c':[1.5]}
    id_order             : list
                           An ordered list of ids, defines the order of
                           observations when iterating over W if not set,
                           lexicographical ordering is used to iterate and the
                           id_order_set property will return False. This can be
                           set after creation by setting the 'id_order' property.
    silence_warnings     : boolean
                           By default libpysal will print a warning if the
                           dataset contains any disconnected components or
                           islands. To silence this warning set this
                           parameter to True.
    ids                  : list
                           Values to use for keys of the neighbors and weights dicts.

    Attributes (NOTE: these are described by their docstrings. to view, use the `help` function)
    ----------

    asymmetries
    cardinalities
    component_labels
    diagW2
    diagWtW
    diagWtW_WW
    histogram
    id2i
    id_order
    id_order_set
    islands
    max_neighbors
    mean_neighbors
    min_neighbors
    n
    n_components
    neighbor_offsets
    nonzero
    pct_nonzero
    s0
    s1
    s2
    s2array
    sd
    sparse
    trcW2
    trcWtW
    trcWtW_WW
    transform

    Examples
    --------
    >>> from libpysal.weights import W
    >>> neighbors = {0: [3, 1], 1: [0, 4, 2], 2: [1, 5], 3: [0, 6, 4], 4: [1, 3, 7, 5], 5: [2, 4, 8], 6: [3, 7], 7: [4, 6, 8], 8: [5, 7]}
    >>> weights = {0: [1, 1], 1: [1, 1, 1], 2: [1, 1], 3: [1, 1, 1], 4: [1, 1, 1, 1], 5: [1, 1, 1], 6: [1, 1], 7: [1, 1, 1], 8: [1, 1]}
    >>> w = W(neighbors, weights)
    >>> "%.3f"%w.pct_nonzero
    '29.630'

    Read from external gal file

    >>> import libpysal
    >>> w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
    >>> w.n
    78
    >>> "%.3f"%w.pct_nonzero
    '6.542'

    Set weights implicitly

    >>> neighbors = {0: [3, 1], 1: [0, 4, 2], 2: [1, 5], 3: [0, 6, 4], 4: [1, 3, 7, 5], 5: [2, 4, 8], 6: [3, 7], 7: [4, 6, 8], 8: [5, 7]}
    >>> w = W(neighbors)
    >>> round(w.pct_nonzero,3)
    29.63
    >>> from libpysal.weights import lat2W
    >>> w = lat2W(100, 100)
    >>> w.trcW2
    39600.0
    >>> w.trcWtW
    39600.0
    >>> w.transform='r'
    >>> round(w.trcW2, 3)
    2530.722
    >>> round(w.trcWtW, 3)
    2533.667

    Cardinality Histogram
    >>> w.histogram
    [(2, 4), (3, 392), (4, 9604)]

    Disconnected observations (islands)

    >>> from libpysal.weights import W
    >>> w = W({1:[0],0:[1],2:[], 3:[]})

    UserWarning: The weights matrix is not fully connected:
    There are 3 disconnected components.
    There are 2 islands with ids: 2, 3.

    """
    def __init__(self,
                 neighbors,
                 weights=None,
                 id_order=None,
                 silence_warnings=False,
                 ids=None):
        self.silence_warnings = silence_warnings
        self.transformations = {}
        self.neighbors = neighbors
        if not weights:
            weights = {}
            for key in neighbors:
                weights[key] = [1.] * len(neighbors[key])
        self.weights = weights
        self.transformations['O'] = self.weights.copy()  # original weights
        self.transform = 'O'
        if id_order is None:
            self._id_order = list(self.neighbors.keys())
            self._id_order.sort()
            self._id_order_set = False
        else:
            self._id_order = id_order
            self._id_order_set = True
        self._reset()
        self._n = len(self.weights)
        if not self.silence_warnings and self.n_components > 1:
            message = "The weights matrix is not fully connected: " \
                          "\n There are %d disconnected components." % \
                          self.n_components
            ni = len(self.islands)
            if ni == 1:
                message = message + "\n There is 1 island with id: " \
                                    "%s."% (str(self.islands[0]))
            elif ni > 1:
                message = message + "\n There are %d islands with ids: %s." % (
                    ni, ', '.join(str(island) for island in self.islands))
            warnings.warn(message)

    def _reset(self):
        """Reset properties.

        """
        self._cache = {}

    @classmethod
    def from_file(cls, path='', format=None, **kwargs):
        f = popen(dataPath=path, mode='r', dataFormat=format)
        w = f.read(**kwargs)
        f.close()
        return w

    @classmethod
    def from_shapefile(cls, *args, **kwargs):
        # we could also just "do the right thing," but I think it'd make sense to
        # try and get people to use `Rook.from_shapefile(shapefile)` rather than
        # W.from_shapefile(shapefile, type=`rook`), otherwise we'd need to build
        # a type dispatch table. Generic W should be for stuff we don't know
        # anything about.
        raise NotImplementedError('Use type-specific constructors, like Rook,'
                                  ' Queen, DistanceBand, or Kernel')

    @classmethod
    def from_WSP(cls, WSP, silence_warnings=True):
        return WSP2W(WSP, silence_warnings=silence_warnings)

    @classmethod
    def from_adjlist(cls,
                     adjlist,
                     focal_col='focal',
                     neighbor_col='neighbor',
                     weight_col=None):
        """
        Return an adjacency list representation of a weights object. 

        Parameters
        ----------
        adjlist         :   pandas DataFrame
                            adjacency list with a minimum of two columns
        focal_col       :   string
                            name of the column with the "source" node ids
        neighbor_col    :   string
                            name of the column with the "destination" node ids
        weight_col      :   string
                            name of the column with the weight information. If not provided and
                            the dataframe has no column named "weight" then all weights
                            are assumed to be 1.
        """
        if weight_col is None:
            weight_col = 'weight'
        try_weightcol = getattr(adjlist, weight_col)
        if try_weightcol is None:
            adjlist = adjlist.copy(deep=True)
            adjlist['weight'] = 1
        all_ids = set(adjlist[focal_col].tolist())
        all_ids |= set(adjlist[neighbor_col].tolist())
        grouper = adjlist.groupby(focal_col)
        neighbors = grouper[neighbor_col].apply(list).to_dict()
        weights = grouper[weight_col].apply(list).to_dict()
        neighbors.update(
            {k: []
             for k in all_ids.difference(list(neighbors.keys()))})
        weights.update(
            {k: []
             for k in all_ids.difference(list(weights.keys()))})
        return cls(neighbors=neighbors, weights=weights)

    def to_adjlist(self,
                   remove_symmetric=False,
                   focal_col='focal',
                   neighbor_col='neighbor',
                   weight_col='weight'):
        """
        Compute an adjacency list representation of a weights object.

        Parameters
        ----------
        remove_symmetric    :   bool
                            whether or not to remove symmetric entries. If the W is symmetric, a standard directed adjacency list will contain both the forward and backward links by default because adjacency lists are a directed graph representation. If this is True, a W created from this adjacency list **MAY NOT BE THE SAME** as the original W. If you would like to consider (1,2) and (2,1) as distinct links, leave this as "False".
        focal_col       :   string
                            name of the column in which to store "source" node ids.
        neighbor_col    :   string
                            name of the column in which to store "destination" node ids.
        weight_col      :   string
                            name of the column in which to store weight information.
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError('pandas must be installed to use this method')
        n_islands = len(self.islands)
        if n_islands > 0 and (not self.silence_warnings):
            warnings.warn('{} islands in this weights matrix. Conversion to an '
                          'adjacency list will drop these observations!')
        adjlist = pd.DataFrame(((idx, n,w) for idx, neighb in self 
                                           for n,w in list(neighb.items())),
                               columns = ('focal', 'neighbor', 'weight'))
        return adjtools.filter_adjlist(adjlist) if remove_symmetric else adjlist

    def to_networkx(self):
        """
        Convert a weights object to a networkx graph

        Parameters
        ----------
        None

        Returns
        -------
        a networkx graph representation of the W object
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError("NetworkX is required to use this function.")
        G = nx.DiGraph() if len(self.asymmetries) > 0 else nx.Graph()
        return nx.from_scipy_sparse_matrix(self.sparse, create_using=G)

    @classmethod
    def from_networkx(cls, graph, weight_col='weight'):
        """
        Convert a networkx graph to a PySAL W object.

        Parameters
        ----------
        graph       :   networkx graph
                        the graph to convert to a W
        weight_col  :   string
                        if the graph is labeled, this should be the
                        name of the field to use as the weight for
                        the W.
        Returns
        --------
        a pysal.weights.W object containing the same graph
        as the networkx graph
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError("NetworkX is required to use this function.")
        sparse_matrix = nx.to_scipy_sparse_matrix(graph)
        return WSP(sparse_matrix).to_W()
        neighbors = dict()
        weights = dict()
        for focal in graph.nodes():
            links = graph[focal]
            neighbors.update({focal: []})
            weights.update({focal: []})
            for neighbor, weight in list(links.items()):
                neighbors[focal].append(neighbor)
                if weight == {}:
                    weights[focal].append(1)
                else:
                    weights[focal].append(weight[weight_col])
        return cls(neighbors=neighbors, weights=weights)

    @property
    def sparse(self):
        """Sparse matrix object.

        For any matrix manipulations required for w, w.sparse should be
        used. This is based on scipy.sparse.

        """
        if 'sparse' not in self._cache:
            self._sparse = self._build_sparse()
            self._cache['sparse'] = self._sparse
        return self._sparse

    @property
    def n_components(self):
        """Store whether the adjacency matrix is fully connected.
        """
        if 'n_components' not in self._cache:
            self._n_components, self._component_labels = connected_components(
                self.sparse)
            self._cache['n_components'] = self._n_components
            self._cache['component_labels'] = self._component_labels
        return self._n_components

    @property
    def component_labels(self):
        """Store the graph component in which each observation falls.
        """
        if 'component_labels' not in self._cache:
            self._n_components, self._component_labels = connected_components(
                self.sparse)
            self._cache['n_components'] = self._n_components
            self._cache['component_labels'] = self._component_labels
        return self._component_labels

    def _build_sparse(self):
        """Construct the sparse attribute.

        """

        row = []
        col = []
        data = []
        id2i = self.id2i
        for i, neigh_list in list(self.neighbor_offsets.items()):
            card = self.cardinalities[i]
            row.extend([id2i[i]] * card)
            col.extend(neigh_list)
            data.extend(self.weights[i])
        row = np.array(row)
        col = np.array(col)
        data = np.array(data)
        s = scipy.sparse.csr_matrix((data, (row, col)), shape=(self.n, self.n))
        return s

    @property
    def id2i(self):
        """Dictionary where the key is an ID and the value is that ID's
        index in W.id_order.

        """
        if 'id2i' not in self._cache:
            self._id2i = {}
            for i, id_i in enumerate(self._id_order):
                self._id2i[id_i] = i
            self._id2i = self._id2i
            self._cache['id2i'] = self._id2i
        return self._id2i

    @property
    def n(self):
        """Number of units.

        """
        if "n" not in self._cache:
            self._n = len(self.neighbors)
            self._cache['n'] = self._n
        return self._n

    @property
    def s0(self):
        """s0 is defined as

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        """
        if 's0' not in self._cache:
            self._s0 = self.sparse.sum()
            self._cache['s0'] = self._s0
        return self._s0

    @property
    def s1(self):
        """s1 is defined as

        .. math::

               s1=1/2 \sum_i \sum_j (w_{i,j} + w_{j,i})^2

        """
        if 's1' not in self._cache:
            t = self.sparse.transpose()
            t = t + self.sparse
            t2 = t.multiply(t)  # element-wise square
            self._s1 = t2.sum() / 2.
            self._cache['s1'] = self._s1
        return self._s1

    @property
    def s2array(self):
        """Individual elements comprising s2.

        See Also
        --------
        s2

        """
        if 's2array' not in self._cache:
            s = self.sparse
            self._s2array = np.array(s.sum(1) + s.sum(0).transpose())**2
            self._cache['s2array'] = self._s2array
        return self._s2array

    @property
    def s2(self):
        """s2 is defined as

        .. math::

                s2=\sum_j (\sum_i w_{i,j} + \sum_i w_{j,i})^2

        """
        if 's2' not in self._cache:
            self._s2 = self.s2array.sum()
            self._cache['s2'] = self._s2
        return self._s2

    @property
    def trcW2(self):
        """Trace of :math:`WW`.

        See Also
        --------
        diagW2

        """
        if 'trcW2' not in self._cache:
            self._trcW2 = self.diagW2.sum()
            self._cache['trcw2'] = self._trcW2
        return self._trcW2

    @property
    def diagW2(self):
        """Diagonal of :math:`WW`.

        See Also
        --------
        trcW2

        """
        if 'diagw2' not in self._cache:
            self._diagW2 = (self.sparse * self.sparse).diagonal()
            self._cache['diagW2'] = self._diagW2
        return self._diagW2

    @property
    def diagWtW(self):
        """Diagonal of :math:`W^{'}W`.

        See Also
        --------
        trcWtW

        """
        if 'diagWtW' not in self._cache:
            self._diagWtW = (self.sparse.transpose() * self.sparse).diagonal()
            self._cache['diagWtW'] = self._diagWtW
        return self._diagWtW

    @property
    def trcWtW(self):
        """Trace of :math:`W^{'}W`.

        See Also
        --------
        diagWtW

        """
        if 'trcWtW' not in self._cache:
            self._trcWtW = self.diagWtW.sum()
            self._cache['trcWtW'] = self._trcWtW
        return self._trcWtW

    @property
    def diagWtW_WW(self):
        """Diagonal of :math:`W^{'}W + WW`.

        """
        if 'diagWtW_WW' not in self._cache:
            wt = self.sparse.transpose()
            w = self.sparse
            self._diagWtW_WW = (wt * w + w * w).diagonal()
            self._cache['diagWtW_WW'] = self._diagWtW_WW
        return self._diagWtW_WW

    @property
    def trcWtW_WW(self):
        """Trace of :math:`W^{'}W + WW`.

        """
        if 'trcWtW_WW' not in self._cache:
            self._trcWtW_WW = self.diagWtW_WW.sum()
            self._cache['trcWtW_WW'] = self._trcWtW_WW
        return self._trcWtW_WW

    @property
    def pct_nonzero(self):
        """Percentage of nonzero weights.

        """
        if 'pct_nonzero' not in self._cache:
            self._pct_nonzero = 100. * self.sparse.nnz / (1. * self._n**2)
            self._cache['pct_nonzero'] = self._pct_nonzero
        return self._pct_nonzero

    @property
    def cardinalities(self):
        """Number of neighbors for each observation.

        """
        if 'cardinalities' not in self._cache:
            c = {}
            for i in self._id_order:
                c[i] = len(self.neighbors[i])
            self._cardinalities = c
            self._cache['cardinalities'] = self._cardinalities
        return self._cardinalities

    @property
    def max_neighbors(self):
        """Largest number of neighbors.

        """
        if 'max_neighbors' not in self._cache:
            self._max_neighbors = max(self.cardinalities.values())
            self._cache['max_neighbors'] = self._max_neighbors
        return self._max_neighbors

    @property
    def mean_neighbors(self):
        """Average number of neighbors.

        """
        if 'mean_neighbors' not in self._cache:
            self._mean_neighbors = np.mean(list(self.cardinalities.values()))
            self._cache['mean_neighbors'] = self._mean_neighbors
        return self._mean_neighbors

    @property
    def min_neighbors(self):
        """Minimum number of neighbors.

        """
        if 'min_neighbors' not in self._cache:
            self._min_neighbors = min(self.cardinalities.values())
            self._cache['min_neighbors'] = self._min_neighbors
        return self._min_neighbors

    @property
    def nonzero(self):
        """Number of nonzero weights.

        """
        if 'nonzero' not in self._cache:
            self._nonzero = self.sparse.nnz
            self._cache['nonzero'] = self._nonzero
        return self._nonzero

    @property
    def sd(self):
        """Standard deviation of number of neighbors.

        """
        if 'sd' not in self._cache:
            self._sd = np.std(list(self.cardinalities.values()))
            self._cache['sd'] = self._sd
        return self._sd

    @property
    def asymmetries(self):
        """List of id pairs with asymmetric weights.

        """
        if 'asymmetries' not in self._cache:
            self._asymmetries = self.asymmetry()
            self._cache['asymmetries'] = self._asymmetries
        return self._asymmetries

    @property
    def islands(self):
        """List of ids without any neighbors.

        """
        if 'islands' not in self._cache:
            self._islands = [
                i for i, c in list(self.cardinalities.items()) if c == 0
            ]
            self._cache['islands'] = self._islands
        return self._islands

    @property
    def histogram(self):
        """Cardinality histogram as a dictionary where key is the id and
        value is the number of neighbors for that unit.

        """
        if 'histogram' not in self._cache:
            ct, bin = np.histogram(
                list(self.cardinalities.values()),
                list(range(self.min_neighbors, self.max_neighbors + 2)))
            self._histogram = list(zip(bin, ct))
            self._cache['histogram'] = self._histogram
        return self._histogram

    def __getitem__(self, key):
        """Allow a dictionary like interaction with the weights class.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w = lat2W()

        >>> w[0] == dict({1: 1.0, 5: 1.0})
        True
        """
        return dict(list(zip(self.neighbors[key], self.weights[key])))

    def __iter__(self):
        """
        Support iteration over weights.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> for i,wi in enumerate(w):
        ...     print(i,wi[0])
        ...
        0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        >>>
        """
        for i in self._id_order:
            yield i, dict(list(zip(self.neighbors[i], self.weights[i])))

    def remap_ids(self, new_ids):
        '''
        In place modification throughout `W` of id values from `w.id_order` to
        `new_ids` in all

        ...

        Parameters
        ----------

        new_ids     :   list
                        /ndarray
                        Aligned list of new ids to be inserted. Note that first
                        element of new_ids will replace first element of
                        w.id_order, second element of new_ids replaces second
                        element of w.id_order and so on.

        Examples
        --------

        >>> from libpysal.weights import lat2W
        >>> w = lat2W(3, 3)
        >>> w.id_order
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> w.neighbors[0]
        [3, 1]
        >>> new_ids = ['id%i'%id for id in w.id_order]
        >>> _ = w.remap_ids(new_ids)
        >>> w.id_order
        ['id0', 'id1', 'id2', 'id3', 'id4', 'id5', 'id6', 'id7', 'id8']
        >>> w.neighbors['id0']
        ['id3', 'id1']
        '''
        old_ids = self._id_order
        if len(old_ids) != len(new_ids):
            raise Exception("W.remap_ids: length of `old_ids` does not match \
            that of new_ids")
        if len(set(new_ids)) != len(new_ids):
            raise Exception("W.remap_ids: list `new_ids` contains duplicates")
        else:
            new_neighbors = {}
            new_weights = {}
            old_transformations = self.transformations['O'].copy()
            new_transformations = {}
            for o, n in zip(old_ids, new_ids):
                o_neighbors = self.neighbors[o]
                o_weights = self.weights[o]
                n_neighbors = [new_ids[old_ids.index(j)] for j in o_neighbors]
                new_neighbors[n] = n_neighbors
                new_weights[n] = o_weights[:]
                new_transformations[n] = old_transformations[o]
            self.neighbors = new_neighbors
            self.weights = new_weights
            self.transformations["O"] = new_transformations

            id_order = [self._id_order.index(o) for o in old_ids]
            for i, id_ in enumerate(id_order):
                self.id_order[id_] = new_ids[i]

            self._reset()

    def __set_id_order(self, ordered_ids):
        """
        Set the iteration order in w.

        W can be iterated over. On construction the iteration order is set to
        the lexicographic order of the keys in the w.weights dictionary. If a specific order
        is required it can be set with this method.

        Parameters
        ----------

        ordered_ids : sequence
                      identifiers for observations in specified order

        Notes
        -----

        ordered_ids is checked against the ids implied by the keys in
        w.weights. If they are not equivalent sets an exception is raised and
        the iteration order is not changed.

        Examples
        --------

        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> for i,wi in enumerate(w):
        ...     print(i, wi[0])
        ...
        0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        >>> w.id_order
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> w.id_order=range(8,-1,-1)
        >>> list(w.id_order)
        [8, 7, 6, 5, 4, 3, 2, 1, 0]
        >>> for i,w_i in enumerate(w):
        ...     print(i,w_i[0])
        ...
        0 8
        1 7
        2 6
        3 5
        4 4
        5 3
        6 2
        7 1
        8 0

        """

        if set(self._id_order) == set(ordered_ids):
            self._id_order = ordered_ids
            self._id_order_set = True
            self._reset()
        else:
            raise Exception('ordered_ids do not align with W ids')

    def __get_id_order(self):
        """Returns the ids for the observations in the order in which they
        would be encountered if iterating over the weights.

        """
        return self._id_order

    id_order = property(__get_id_order, __set_id_order)

    @property
    def id_order_set(self):
        """
        Returns True if user has set id_order, False if not.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.id_order_set
        True
        """
        return self._id_order_set

    @property
    def neighbor_offsets(self):
        """
        Given the current id_order, neighbor_offsets[id] is the offsets of the
        id's neighbors in id_order.

        Returns
        -------
        list
                offsets of the id's neighbors in id_order

        Examples
        --------
        >>> from libpysal.weights import W
        >>> neighbors={'c': ['b'], 'b': ['c', 'a'], 'a': ['b']}
        >>> weights ={'c': [1.0], 'b': [1.0, 1.0], 'a': [1.0]}
        >>> w=W(neighbors,weights)
        >>> w.id_order = ['a','b','c']
        >>> w.neighbor_offsets['b']
        [2, 0]
        >>> w.id_order = ['b','a','c']
        >>> w.neighbor_offsets['b']
        [2, 1]
        """

        if "neighbors_0" not in self._cache:
            self.__neighbors_0 = {}
            id2i = self.id2i
            for j, neigh_list in list(self.neighbors.items()):
                self.__neighbors_0[j] = [id2i[neigh] for neigh in neigh_list]
            self._cache['neighbors_0'] = self.__neighbors_0
        return self.__neighbors_0

    def get_transform(self):
        """
        Getter for transform property.

        Returns
        -------
        transformation : string (or none)

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.weights[0]
        [1.0, 1.0]
        >>> w.transform
        'O'
        >>> w.transform='r'
        >>> w.weights[0]
        [0.5, 0.5]
        >>> w.transform='b'
        >>> w.weights[0]
        [1.0, 1.0]


        See also
        --------
        set_transform


        """

        return self._transform

    def set_transform(self, value="B"):
        """
        Transformations of weights.

        Notes
        -----

        Transformations are applied only to the value of the weights at
        instantiation. Chaining of transformations cannot be done on a W
        instance.

        Parameters
        ----------
        transform   :   string
                        not case sensitive)

        .. table::

           :widths: auto

           ================   ======================================================
           transform string   value
           ================   ======================================================
           B                  Binary
           R                  Row-standardization (global sum=n)
           D                  Double-standardization (global sum=1)
           V                  Variance stabilizing
           O                  Restore original transformation (from instantiation)
           ================   ======================================================

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.weights[0]
        [1.0, 1.0]
        >>> w.transform
        'O'
        >>> w.transform='r'
        >>> w.weights[0]
        [0.5, 0.5]
        >>> w.transform='b'
        >>> w.weights[0]
        [1.0, 1.0]
        """
        value = value.upper()
        self._transform = value
        if value in self.transformations:
            self.weights = self.transformations[value]
            self._reset()
        else:
            if value == "R":
                # row standardized weights
                weights = {}
                self.weights = self.transformations['O']
                for i in self.weights:
                    wijs = self.weights[i]
                    row_sum = sum(wijs) * 1.0
                    if row_sum == 0.0:
                        if not self.silence_warnings:
                            print(('WARNING: ', i,
                                   ' is an island (no neighbors)'))
                    weights[i] = [wij / row_sum for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "D":
                # doubly-standardized weights
                # update current chars before doing global sum
                self._reset()
                s0 = self.s0
                ws = 1.0 / s0
                weights = {}
                self.weights = self.transformations['O']
                for i in self.weights:
                    wijs = self.weights[i]
                    weights[i] = [wij * ws for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "B":
                # binary transformation
                weights = {}
                self.weights = self.transformations['O']
                for i in self.weights:
                    wijs = self.weights[i]
                    weights[i] = [1.0 for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "V":
                # variance stabilizing
                weights = {}
                q = {}
                k = self.cardinalities
                s = {}
                Q = 0.0
                self.weights = self.transformations['O']
                for i in self.weights:
                    wijs = self.weights[i]
                    q[i] = math.sqrt(sum([wij * wij for wij in wijs]))
                    s[i] = [wij / q[i] for wij in wijs]
                    Q += sum([si for si in s[i]])
                nQ = self.n / Q
                for i in self.weights:
                    weights[i] = [w * nQ for w in s[i]]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "O":
                # put weights back to original transformation
                weights = {}
                original = self.transformations[value]
                self.weights = original
                self._reset()
            else:
                raise Exception('unsupported weights transformation')

    transform = property(get_transform, set_transform)

    def asymmetry(self, intrinsic=True):
        """
        Asymmetry check.

        Parameters
        ----------
        intrinsic   :   boolean
                        default=True

                intrinsic symmetry:
                      :math:`w_{i,j} == w_{j,i}`

                if intrisic is False:
                    symmetry is defined as :math:`i \in N_j \ AND \ j \in N_i` where
                    :math:`N_j` is the set of neighbors for j.

        Returns
        -------
        asymmetries : list
                      empty if no asymmetries are found
                      if asymmetries, then a list of (i,j) tuples is returned

        Examples
        --------

        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> w.asymmetry()
        []
        >>> w.transform='r'
        >>> w.asymmetry()
        [(0, 1), (0, 3), (1, 0), (1, 2), (1, 4), (2, 1), (2, 5), (3, 0), (3, 4), (3, 6), (4, 1), (4, 3), (4, 5), (4, 7), (5, 2), (5, 4), (5, 8), (6, 3), (6, 7), (7, 4), (7, 6), (7, 8), (8, 5), (8, 7)]
        >>> result = w.asymmetry(intrinsic=False)
        >>> result
        []
        >>> neighbors={0:[1,2,3], 1:[1,2,3], 2:[0,1], 3:[0,1]}
        >>> weights={0:[1,1,1], 1:[1,1,1], 2:[1,1], 3:[1,1]}
        >>> w=W(neighbors,weights)
        >>> w.asymmetry()
        [(0, 1), (1, 0)]
        """

        if intrinsic:
            wd = self.sparse.transpose() - self.sparse
        else:
            transform = self.transform
            self.transform = 'b'
            wd = self.sparse.transpose() - self.sparse
            self.transform = transform

        ids = np.nonzero(wd)
        if len(ids[0]) == 0:
            return []
        else:
            ijs = list(zip(ids[0], ids[1]))
            ijs.sort()
            return ijs

    def symmetrize(self, inplace=False):
        """
        Construct a symmetric KNN weight.

        This ensures that the neighbors of each focal observation
        consider the focal observation itself as a neighbor. 

        This returns a generic W object, since the object is no
        longer guaranteed to have k neighbors for each observation.
        """
        if not inplace:
            neighbors = copy.deepcopy(self.neighbors)
            weights = copy.deepcopy(self.weights)
            out_W = W(neighbors, weights, id_order=self.id_order)
            out_W.symmetrize(inplace=True)
            return out_W
        else:
            for focal, fneighbs in list(self.neighbors.items()):
                for j, neighbor in enumerate(fneighbs):
                    neighb_neighbors = self.neighbors[neighbor]
                    if focal not in neighb_neighbors:
                        self.neighbors[neighbor].append(focal)
                        self.weights[neighbor].append(self.weights[focal][j])
            self._cache = dict()
            return

    def full(self):
        """
        Generate a full numpy array.

        Parameters
        ----------
        self        : W
                   spatial weights object

        Returns
        -------
        (fullw, keys) : tuple
                        first element being the full numpy array and second element
                        keys being the ids associated with each row in the array.

        Examples
        --------
        >>> from libpysal.weights import W, full
        >>> neighbors = {'first':['second'],'second':['first','third'],'third':['second']}
        >>> weights = {'first':[1],'second':[1,1],'third':[1]}
        >>> w = W(neighbors, weights)
        >>> wf, ids = full(w)
        >>> wf
        array([[0., 1., 0.],
               [1., 0., 1.],
               [0., 1., 0.]])
        >>> ids
        ['first', 'second', 'third']
        """
        wfull = np.zeros([self.n, self.n], dtype=float)
        keys = list(self.neighbors.keys())
        if self.id_order:
            keys = self.id_order
        for i, key in enumerate(keys):
            n_i = self.neighbors[key]
            w_i = self.weights[key]
            for j, wij in zip(n_i, w_i):
                c = keys.index(j)
                wfull[i, c] = wij
        return (wfull, keys)

    def to_WSP(self):
        '''
        Generate a WSP object.

        Returns
        -------

        implicit : libpysal.weights.WSP
                   Thin W class

        Examples
        --------
        >>> from libpysal.weights import W, WSP
        >>> neighbors={'first':['second'],'second':['first','third'],'third':['second']}
        >>> weights={'first':[1],'second':[1,1],'third':[1]}
        >>> w=W(neighbors,weights)
        >>> wsp=w.to_WSP()
        >>> isinstance(wsp, WSP)
        True
        >>> wsp.n
        3
        >>> wsp.s0
        4

        See also
        --------
        WSP

        '''
        return WSP(self.sparse, self._id_order)

    def set_shapefile(self, shapefile, idVariable=None, full=False):
        """
        Adding meta data for writing headers of gal and gwt files.

        Parameters
        ----------

        shapefile :     string
                        shapefile name used to construct weights

        idVariable :    string
                        name of attribute in shapefile to associate with ids in the weights

        full :          boolean
                        True - write out entire path for shapefile, False
                        (default) only base of shapefile without extension

        """

        if full:
            self._shpName = shapefile
        else:
            self._shpName = BASENAME(shapefile).split(".")[0]

        self._varName = idVariable

    def plot(self,
             gdf,
             indexed_on=None,
             ax=None,
             color='k',
             node_kws=None,
             edge_kws=None):
        """
        Plot spatial weights objects.
        NOTE: Requires matplotlib, and implicitly requires geopandas 
        dataframe as input.

        Parameters
        ----------
        gdf         : geopandas geodataframe 
                      the original shapes whose topological relations are 
                      modelled in W. 
        indexed_on  : str 
                      column of gdf which the weights object uses as an index.
                      (Default: None, so the geodataframe's index is used)
        ax          : matplotlib axis
                      axis on which to plot the weights. 
                      (Default: None, so plots on the current figure)
        color       : string
                      matplotlib color string, will color both nodes and edges
                      the same by default. 
        node_kws    : keyword argument dictionary
                      dictionary of keyword arguments to send to pyplot.scatter,
                      which provide fine-grained control over the aesthetics
                      of the nodes in the plot
        edge_kws    : keyword argument dictionary
                      dictionary of keyword arguments to send to pyplot.plot,
                      which provide fine-grained control over the aesthetics
                      of the edges in the plot

        Returns
        -------
        f,ax : matplotlib figure,axis on which the plot is made. 

        NOTE: if you'd like to overlay the actual shapes from the 
              geodataframe, call gdf.plot(ax=ax) after this. To plot underneath,
              adjust the z-order of the geopandas plot: gdf.plot(ax=ax,zorder=0)

        Examples
        --------

        >>> from libpysal.weights import Queen
        >>> import libpysal as lp
        >>> import geopandas
        >>> gdf = geopandas.read_file(lp.examples.get_path("columbus.shp"))
        >>> weights = Queen.from_dataframe(gdf)
        >>> tmp = weights.plot(gdf, color='firebrickred', node_kws=dict(marker='*', color='k'))
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError(
                "W.plot depends on matplotlib.pyplot, and this was"
                "not able to be imported. \nInstall matplotlib to"
                "plot spatial weights.")
        if ax is None:
            f = plt.figure()
            ax = plt.gca()
        else:
            f = plt.gcf()
        if node_kws is not None:
            if 'color' not in node_kws:
                node_kws['color'] = color
        else:
            node_kws = dict(color=color)
        if edge_kws is not None:
            if 'color' not in edge_kws:
                edge_kws['color'] = color
        else:
            edge_kws = dict(color=color)

        for idx, neighbors in self:
            if idx in self.islands:
                continue
            if indexed_on is not None:
                neighbors = gdf[gdf[indexed_on].isin(neighbors)].index.tolist()
                idx = gdf[gdf[indexed_on] == idx].index.tolist()[0]
            centroids = gdf.loc[neighbors].centroid.apply(lambda p: (p.x, p.y))
            centroids = np.vstack(centroids.values)
            focal = np.hstack(gdf.loc[idx].geometry.centroid.xy)
            seen = set()
            for nidx, neighbor in zip(neighbors, centroids):
                if (idx, nidx) in seen:
                    continue
                ax.plot(*list(zip(focal, neighbor)), marker=None, **edge_kws)
                seen.update((idx, nidx))
                seen.update((nidx, idx))
        ax.scatter(gdf.centroid.apply(lambda p: p.x),
                   gdf.centroid.apply(lambda p: p.y), **node_kws)
        return f, ax


class WSP(object):
    """
    Thin W class for spreg.

    Parameters
    ----------

    sparse   : sparse_matrix
               NxN object from scipy.sparse

    id_order : list
               An ordered list of ids, assumed to match the ordering in
               sparse.

    Attributes
    ----------

    n           : int
                  description
    s0          : float
                  description
    trcWtW_WW   : float
                  description

    Examples
    --------

    From GAL information

    >>> import scipy.sparse
    >>> from libpysal.weights import WSP
    >>> rows = [0, 1, 1, 2, 2, 3]
    >>> cols = [1, 0, 2, 1, 3, 3]
    >>> weights =  [1, 0.75, 0.25, 0.9, 0.1, 1]
    >>> sparse = scipy.sparse.csr_matrix((weights, (rows, cols)), shape=(4,4))
    >>> w = WSP(sparse)
    >>> w.s0
    4.0
    >>> w.trcWtW_WW
    6.395
    >>> w.n
    4

    """
    def __init__(self, sparse, id_order=None):
        if not scipy.sparse.issparse(sparse):
            raise ValueError("must pass a scipy sparse object")
        rows, cols = sparse.shape
        if rows != cols:
            raise ValueError("Weights object must be square")
        self.sparse = sparse.tocsr()
        self.n = sparse.shape[0]
        if id_order:
            if len(id_order) != self.n:
                raise ValueError(
                    "Number of values in id_order must match shape of sparse")
        self.id_order = id_order
        self._cache = {}

    @property
    def s0(self):
        """s0 is defined as:

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        """
        if 's0' not in self._cache:
            self._s0 = self.sparse.sum()
            self._cache['s0'] = self._s0
        return self._s0

    @property
    def trcWtW_WW(self):
        """Trace of :math:`W^{'}W + WW`.

        """
        if 'trcWtW_WW' not in self._cache:
            self._trcWtW_WW = self.diagWtW_WW.sum()
            self._cache['trcWtW_WW'] = self._trcWtW_WW
        return self._trcWtW_WW

    @property
    def diagWtW_WW(self):
        """Diagonal of :math:`W^{'}W + WW`.

        """
        if 'diagWtW_WW' not in self._cache:
            wt = self.sparse.transpose()
            w = self.sparse
            self._diagWtW_WW = (wt * w + w * w).diagonal()
            self._cache['diagWtW_WW'] = self._diagWtW_WW
        return self._diagWtW_WW

    @classmethod
    def from_W(cls, W):
        """
        Constructs a WSP object from the W's sparse matrix

        Parameters
        ----------
        W       :   libpysal.weights.W
                    a pysal weights object with a sparse form and ids

        Returns
        -------
        a WSP instance
        """
        return cls(W.sparse, id_order=W.id_order)

    def to_W(self, silence_warnings=False):
        """
        Convert a pysal WSP object (thin weights matrix) to a pysal W object.

        Parameters
        ----------
        self                     : WSP
                                  PySAL sparse weights object
        silence_warnings         : boolean
                                  Switch to turn off (default on) print statements
                                  for every observation with islands

        Returns
        -------
        w       : W
                  PySAL weights object

        Examples
        --------
        >>> from libpysal.weights import lat2SW, WSP, WSP2W

        Build a 10x10 scipy.sparse matrix for a rectangular 2x5 region of cells
        (rook contiguity), then construct a libpysal sparse weights object (self).

        >>> sp = lat2SW(2, 5)
        >>> self = WSP(sp)
        >>> self.n
        10
        >>> print(self.sparse[0].todense())
        [[0 1 0 0 0 1 0 0 0 0]]

        Convert this sparse weights object to a standard PySAL weights object.

        >>> w = WSP2W(self)
        >>> w.n
        10
        >>> print(w.full()[0][0])
        [0. 1. 0. 0. 0. 1. 0. 0. 0. 0.]

        """

        indices = list(self.sparse.indices)
        data = list(self.sparse.data)
        indptr = list(self.sparse.indptr)
        id_order = self.id_order
        if id_order:
            # replace indices with user IDs
            indices = [id_order[i] for i in indices]
        else:
            id_order = list(range(self.n))
        neighbors, weights = {}, {}
        start = indptr[0]
        for i in range(self.n):
            oid = id_order[i]
            end = indptr[i + 1]
            neighbors[oid] = indices[start:end]
            weights[oid] = data[start:end]
            start = end
        ids = copy.copy(self.id_order)
        w = W(neighbors, weights, ids, silence_warnings=silence_warnings)
        w._sparse = copy.deepcopy(self.sparse)
        w._cache['sparse'] = w._sparse
        return w



"""
Voronoi tesslation of 2-d point sets


Adapted from https://gist.github.com/pv/8036995

"""
import numpy as np
from scipy.spatial import Voronoi

__author__ = "Serge Rey <sjsrey@gmail.com>"

__all__ = ['voronoi_frames']


def voronoi(points, radius=None):
    """
    Determine finite Voronoi diagram for a 2-d point set 


    Parameters
    ----------
    points      : array-like
                  nx2 array of points

    radius      : float (optional) 
                  distance to 'points at infinity'

    Returns
    -------

    regions    : list
                  each element of the list contains sequence of the indexes of Voronoi vertices composing a Voronoi polygon (region)

    coordinates : array
                  coordinates of the Voronoi vertices


    Examples
    --------
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions, coordinates = voronoi(points)
    >>> regions
    [[1, 3, 2], [4, 5, 1, 0], [0, 1, 7, 6], [9, 0, 8]]
    >>> coordinates
    array([[  4.21783296,   4.08408578],
           [  7.51956025,   3.51807539],
           [  9.4642193 ,  19.3994576 ],
           [ 14.98210684, -10.63503022],
           [ -9.22691341,  -4.58994414],
           [ 14.98210684, -10.63503022],
           [  1.78491801,  19.89803294],
           [  9.4642193 ,  19.3994576 ],
           [  1.78491801,  19.89803294],
           [ -9.22691341,  -4.58994414]])
    """
    vor = Voronoi(points)
    return voronoi_regions(vor, radius=radius)


def voronoi_regions(vor, radius=None):
    """
    Finite voronoi regions for a 2-d point set.


    Parameters
    ----------

    vor:  Voronoi (scipy.spatial)

    radius: float (optional)
            Distance to 'points at infinity'
    """
    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max() * 2

    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            new_regions.append(vertices)
            continue

        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                continue

            t = vor.points[p2] - vor.points[p1]
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)


def as_dataframes(regions, vertices, points):
    """
    Helper function to store finite Voronoi regions and originator points as
    geopandas (or pandas) data frames


    Parameters
    ----------

    regions     : list
                  each element of the list contains sequence of the indexes of
                  voronoi vertices composing a vornoi polygon (region)

    vertices    : array
                  coordinates of the vornoi vertices

    points      : array-like
                  originator points


    Returns
    -------

    region_df   : GeoDataFrame
                  Finite Voronoi polygons as geometries

    points_df   : GeoDataFrame
                  Originator points as geometries
    """
    try:
        import geopandas as gpd
    except ImportError:
        gpd = None

    try:
        from shapely.geometry import Polygon, Point
    except ImportError:
        from .shapes import Polygon, Point

    if gpd is not None:
        region_df = gpd.GeoDataFrame()
        region_df['geometry'] = [
            Polygon(vertices[region]) for region in regions
        ]

        point_df = gpd.GeoDataFrame()
        point_df['geometry'] = gpd.GeoSeries(Point(pnt) for pnt in points)
    else:
        import pandas as pd
        region_df = pd.DataFrame()
        region_df['geometry'] = [
            Polygon(vertices[region].tolist()) for region in regions
        ]
        point_df = pd.DataFrame()
        point_df['geometry'] = [Point(pnt) for pnt in points]

    return region_df, point_df


def voronoi_frames(points, radius=None, clip='extent'):
    """
    Composite helper to return Voronoi regions and generator points as individual dataframes

    Parameters
    ----------

    points      : array-like
                  originator points

    radius      : float
                  distance to "points at infinity" used in building voronoi cells

    clip        : str, or shapely.geometry.Polygon
                  an overloaded option about how to clip the voronoi
                  cells. The options are:
                  - 'none'/None: no clip is applied. Voronoi cells may be arbitrarily larger that the source map. Note that this may lead to cells that are many orders of magnitude larger in extent than the original map. Not recommended.
                  - 'bbox'/'extent'/'bounding box': clip the voronoi cells to the bounding box of the input points.
                  - 'chull'/'convex hull': clip the voronoi cells to the convex hull of the input points.
                  - 'ashape'/'ahull': clip the voronoi cells to the tightest hull that contains all points (e.g. the smallest alphashape, using libpysal.cg.alpha_shape_auto)
                  - Polygon: clip to an arbitrary Polygon

    tolerance   : float
                  percent of map width to use to buffer the extent of the map, if clipping (default: .01, or 1 percent)

    Returns
    -------

    _           : tuple
                  (region_df, points_df)

                  region_df   : GeoDataFrame (if geopandas available, otherwise Pandas DataFrame)
                                Finite Voronoi polygons as geometries

                  points_df   : GeoDataFrame (if geopandas available, otherwise Pandas DataFrame)
                                Originator points as geometries

    Notes
    -----

    If Geopandas is not available the return types will be Pandas DataFrames
    each with a geometry column populated with PySAL shapes. If Geopandas is
    available, return types are GeoDataFrames with a geometry column populated
    with shapely geometry types.

    Examples
    --------
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions_df, points_df = voronoi_frames(points)
    >>> regions_df.shape
    (4, 1)
    >>> regions_df.shape == points_df.shape
    True

    """
    regions, vertices = voronoi(points, radius=radius)
    regions, vertices = as_dataframes(regions, vertices, points)
    if clip:
        regions = clip_voronoi_frames_to_extent(regions, vertices, clip=clip)
    return regions, vertices


def clip_voronoi_frames_to_extent(regions, vertices, clip='extent'):
    """
    Arguments
    ---------
    regions :   geopandas geodataframe
                dataframe containing voronoi cells to clip
    vertices:   geopandas geodataframe
                dataframe containing vertices used to build voronoi cells
    clip        : str, or shapely.geometry.Polygon
                  an overloaded option about how to clip the voronoi
                  cells. The options are:
                  - 'none'/None: no clip is applied. Voronoi cells may be arbitrarily
                                 larger that the source map. Note that this may lead
                                 to cells that are many orders of magnitude larger
                                 in extent than the original map. Not recommended. 
                  - 'bbox'/'extent'/'bounding box': clip the voronoi cells to the 
                                                    bounding box of the input points.
                  - 'chull'/'convex hull': clip the voronoi cells to the convex hull
                                           of the input points.
                  - 'ashape'/'ahull': clip the voronoi cells to the tightest hull that
                                      contains all points (e.g. the smallest alphashape,
                                      using libpysal.cg.alpha_shape_auto)
                  - Polygon: clip to an arbitrary shapely.geometry.Polygon
    
    Returns
    -------
    a geodataframe of clipped voronoi regions

    """
    try:
        from shapely.geometry import Polygon
    except ImportError:
        raise ImportError("shapely is required to clip voronoi regions")
    try:
        import geopandas
    except ImportError:
        raise ImportError('geopandas is required to clip voronoi regions')

    if isinstance(clip, Polygon):
        clipper = geopandas.GeoDataFrame(geometry=[clip])
    elif clip is None:
        return regions
    elif clip.lower() == 'none':
        return regions
    elif clip.lower() in ('bounds', 'bounding box', 'bbox', 'extent'):
        min_x, min_y, max_x, max_y = vertices.total_bounds
        bounding_poly = Polygon([(min_x, min_y), (min_x, max_y),
                                 (max_x, max_y), (max_x, min_y),
                                 (min_x, min_y)])
        clipper = geopandas.GeoDataFrame(geometry=[bounding_poly])
    elif clip.lower() in ('chull', 'convex hull', 'convex_hull'):
        clipper = geopandas.GeoDataFrame(geometry=[vertices.geometry\
                                                           .unary_union\
                                                           .convex_hull])
    elif clip.lower() in ('ahull', 'alpha hull', 'alpha_hull', 'ashape',
                          'alpha shape', 'alpha_shape'):
        from .alpha_shapes import alpha_shape_auto
        from ..weights.distance import get_points_array
        coordinates = get_points_array(vertices.geometry)
        clipper = geopandas.GeoDataFrame(
            geometry=[alpha_shape_auto(coordinates)])
    else:
        raise ValueError('clip type "{}" not understood. Try one '
                         ' of the supported options: [None, "extent", '
                         '"chull", "ahull"]'.format(clip))
    clipped_regions = geopandas.overlay(regions, clipper, how='intersection')
    return clipped_regions



from ..fileio import FileIO
from ...examples import get_path 
def test_by_col_exists():
    """ 
    Test if the Metaclass is initializing and providing
        readers to its children.
    """
    fh1 = FileIO.open(get_path("columbus.dbf"))
    fh2 = FileIO.open(get_path("usjoin.csv"))

    assert hasattr(fh1, 'by_col')
    assert hasattr(fh2, 'by_col')




import os
import unittest
from ..weights import W
from ..util import lat2W
from ..spatial_lag import lag_spatial, lag_categorical
import numpy as np


class Test_spatial_lag(unittest.TestCase):
    def setUp(self):
        self.neighbors = {'c': ['b'], 'b': ['c', 'a'], 'a': ['b']}
        self.weights = {'c': [1.0], 'b': [1.0, 1.0], 'a': [1.0]}
        self.id_order = ['a', 'b', 'c']
        self.weights = {'c': [1.0], 'b': [1.0, 1.0], 'a': [1.0]}
        self.w = W(self.neighbors, self.weights, self.id_order)
        self.y = np.array([0, 1, 2])
        self.wlat = lat2W(3, 3)
        self.ycat = ['a','b','a','b','c','b','c','b','c']
        self.ycat2 = ['a', 'c', 'c', 'd', 'b', 'a', 'd', 'd', 'c']
        self.ym = np.vstack((self.ycat,self.ycat2)).T
        self.random_seed = 503
    def test_lag_spatial(self):
        yl = lag_spatial(self.w, self.y)
        np.testing.assert_array_almost_equal(yl, [1., 2., 1.])
        self.w.id_order = ['b', 'c', 'a']
        y = np.array([1, 2, 0])
        yl = lag_spatial(self.w, y)
        np.testing.assert_array_almost_equal(yl, [2., 1., 1.])
        w = lat2W(3, 3)
        y = np.arange(9)
        yl = lag_spatial(w, y)
        ylc = np.array([4., 6., 6., 10., 16., 14., 10., 18., 12.])
        np.testing.assert_array_almost_equal(yl, ylc)
        w.transform = 'r'
        yl = lag_spatial(w, y)
        ylc = np.array(
            [2., 2., 3., 3.33333333, 4.,
             4.66666667, 5., 6., 6.])
        np.testing.assert_array_almost_equal(yl, ylc)
    
    def test_lag_categorical(self):
        yl = lag_categorical(self.wlat, self.ycat)
        np.random.seed(self.random_seed)
        known = np.array(['b', 'a', 'b', 'c', 'b', 'c', 'b', 'c', 'b'])
        np.testing.assert_array_equal(yl, known)
        ym_lag = lag_categorical(self.wlat,self.ym)
        known = np.array([['b', 'c'],
                          ['a', 'c'],
                          ['b', 'c'],
                          ['c', 'd'],
                          ['b', 'd'],
                          ['c', 'c'],
                          ['b', 'd'],
                          ['c', 'd'],
                          ['b', 'd']])
        np.testing.assert_array_equal(ym_lag, np.asarray(known))

suite = unittest.TestLoader().loadTestsFromTestCase(Test_spatial_lag)

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite)



import unittest as ut
import numpy as np
from .. import adjtools as adj
from ... import weights
from ... import io
from ... import examples
from ..util import lat2W
from ...common import RTOL, ATOL


try:
    import pandas
    import geopandas
    PANDAS_MISSING = False
except ImportError:
    PANDAS_MISSING = True

@ut.skipIf(PANDAS_MISSING, 'Pandas is gone')
class Test_Adjlist(ut.TestCase):
    def setUp(self):
        self.knownW = io.open(examples.get_path('columbus.gal')).read()

    def test_round_trip(self):
        adjlist = self.knownW.to_adjlist(remove_symmetric=False).astype(int)
        w_from_adj = weights.W.from_adjlist(adjlist)
        np.testing.assert_allclose(w_from_adj.sparse.toarray(),
                                   self.knownW.sparse.toarray())

    def test_filter(self):
        grid = lat2W(2,2)
        alist = grid.to_adjlist(remove_symmetric=True)
        assert len(alist) == 4
        with self.assertRaises(AssertionError):
            badgrid = weights.W.from_adjlist(alist)
            np.testing.assert_allclose(badgrid.sparse.toarray(),
                                       grid.sparse.toarray())
        assert set(alist.focal.unique().tolist()) == set(list(range(4)))
        assert set(alist.neighbor.unique().tolist()) == set(list(range(4)))
        assert alist.weight.unique().item() == 1
        grid = lat2W(2,2, id_type='string')
        alist = grid.to_adjlist(remove_symmetric=True)
        assert len(alist) == 4
        with self.assertRaises(AssertionError):
            badgrid = weights.W.from_adjlist(alist)
            np.testing.assert_allclose(badgrid.sparse.toarray(),
                                       grid.sparse.toarray())
        print(alist)
        tuples = set([tuple(t) for t in alist[['focal','neighbor']].values])
        full_alist = grid.to_adjlist()
        all_possible = set([tuple(t) for t in full_alist[['focal','neighbor']].values])
        assert tuples.issubset(all_possible), ('the de-duped adjlist has links '
                                               'not in the duplicated adjlist.')
        complements = all_possible.difference(tuples)
        reversed_complements = set([t[::-1] for t in complements])
        assert reversed_complements == tuples, ('the remaining links in the duplicated'
                                                ' adjlist are not the reverse of the links'
                                                ' in the deduplicated adjlist.')
        assert alist.weight.unique().item() == 1


    def apply_and_compare_columbus(self, col):
        import geopandas
        df = geopandas.read_file(examples.get_path('columbus.dbf')).head()
        W = weights.Queen.from_dataframe(df)
        alist = adj.adjlist_apply(df[col], W=W)
        right_hovals = alist.groupby('focal').att_focal.unique()
        assert (right_hovals == df[col]).all()
        allpairs = np.subtract.outer(df[col].values, df[col].values)
        flat_diffs = allpairs[W.sparse.toarray().astype(bool)]
        np.testing.assert_allclose(flat_diffs, alist['subtract'].values)
        return flat_diffs

    def test_apply(self):
        self.apply_and_compare_columbus('HOVAL')

    def test_mvapply(self):
        import geopandas
        df = geopandas.read_file(examples.get_path('columbus.dbf')).head()
        W = weights.Queen.from_dataframe(df)
        ssq = lambda x_y: np.sum((x_y[0]-x_y[1])**2).item()
        ssq.__name__ = 'sum_of_squares'
        alist = adj.adjlist_apply(df[['HOVAL', 'CRIME', 'INC']], W=W,
                                  func=ssq)
        known_ssq = [1301.1639302990804,
                     3163.46450914361,
                     1301.1639302990804,
                     499.52656498472993,
                     594.518273032036,
                     3163.46450914361,
                     499.52656498472993,
                     181.79100173844196,
                     436.09336916344097,
                     594.518273032036,
                     181.79100173844196,
                     481.89443401250094,
                     436.09336916344097,
                     481.89443401250094] #ugh I hate doing this, but how else?
        np.testing.assert_allclose(alist.sum_of_squares.values,
                                   np.asarray(known_ssq),
                                   rtol=RTOL, atol=ATOL)

    def test_map(self):
        atts = ['HOVAL', 'CRIME', 'INC']
        df = geopandas.read_file(examples.get_path('columbus.dbf')).head()
        W = weights.Queen.from_dataframe(df)
        hoval, crime, inc = list(map(self.apply_and_compare_columbus, atts))
        mapped = adj.adjlist_map(df[atts], W=W)
        for name,data in zip(atts, (hoval, crime, inc)):
            np.testing.assert_allclose(data,
                                       mapped['_'.join(('subtract',name))].values)



import unittest
from ..arcgis_swm import ArcGISSwmIO
from ...fileio import FileIO as psopen
from .... import examples as pysal_examples
import tempfile
import os


class test_ArcGISSwmIO(unittest.TestCase):
    def setUp(self):
        self.test_file = test_file = pysal_examples.get_path('ohio.swm')
        self.obj = ArcGISSwmIO(test_file, 'r')

    def test_close(self):
        f = self.obj
        f.close()
        self.assertRaises(ValueError, f.read)

    def test_read(self):
        w = self.obj.read()
        self.assertEqual(88, w.n)
        self.assertEqual(5.25, w.mean_neighbors)
        self.assertEqual([1.0, 1.0, 1.0, 1.0], list(w[1].values()))

    def test_seek(self):
        self.test_read()
        self.assertRaises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    def test_write(self):
        w = self.obj.read()
        f = tempfile.NamedTemporaryFile(
            suffix='.swm', dir=pysal_examples.get_path(''))
        fname = f.name
        f.close()
        o = psopen(fname, 'w')
        o.write(w)
        o.close()
        wnew = psopen(fname, 'r').read()
        self.assertEqual(wnew.pct_nonzero, w.pct_nonzero)
        os.remove(fname)

if __name__ == '__main__':
    unittest.main()



import unittest as ut
import numpy as np
from ..util import lat2W
from ..weights import W
try:
    import networkx as nx
except ImportError:
    nx = None

@ut.skipIf(nx is None, "Missing networkx")
class Test_NetworkXConverter(ut.TestCase):
    def setUp(self):
        self.known_nx = nx.random_regular_graph(4,10,seed=8879) 
        self.known_amat = np.array(nx.to_numpy_matrix(self.known_nx))
        self.known_W = lat2W(5,5)

    def test_round_trip(self):
        W_ = W.from_networkx(self.known_nx)
        np.testing.assert_allclose(W_.sparse.toarray(), self.known_amat)
        nx2 = W_.to_networkx()
        np.testing.assert_allclose(nx.to_numpy_matrix(nx2), self.known_amat)
        nxsquare = self.known_W.to_networkx()
        np.testing.assert_allclose(self.known_W.sparse.toarray(), nx.to_numpy_matrix(nxsquare))
        W_square = W.from_networkx(nxsquare)
        np.testing.assert_allclose(self.known_W.sparse.toarray(), W_square.sparse.toarray())



import unittest

import sys
#import pysal_examples
from .... import examples as  pysal_examples
PY3 = int(sys.version[0]) > 2
if PY3:
    import io
else:
    from cStringIO import StringIO

from ..shapefile import noneMax, noneMin, shp_file, shx_file, NullShape, Point, PolyLine, MultiPoint, PointZ, PolyLineZ, PolygonZ, MultiPointZ, PointM, PolyLineM, PolygonM, MultiPointM, MultiPatch
import os

def bufferIO(buf):
    """
    Temp stringIO function to force compat
    """
    if PY3:
        return io.BytesIO(buf)
    else:
        return StringIO(buf)

class TestNoneMax(unittest.TestCase):
    def test_none_max(self):
        self.assertEqual(5, noneMax(5, None))
        self.assertEqual(1, noneMax(None, 1))
        self.assertEqual(None, noneMax(None, None))


class TestNoneMin(unittest.TestCase):
    def test_none_min(self):
        self.assertEqual(5, noneMin(5, None))
        self.assertEqual(1, noneMin(None, 1))
        self.assertEqual(None, noneMin(None, None))


class test_shp_file(unittest.TestCase):
    def test___init__(self):
        shp = shp_file(pysal_examples.get_path('10740.shp'))
        assert shp.header == {'BBOX Xmax': -105.29012, 'BBOX Ymax': 36.219799000000002, 'BBOX Mmax': 0.0, 'BBOX Zmin': 0.0, 'BBOX Mmin': 0.0, 'File Code': 9994, 'BBOX Ymin': 34.259672000000002, 'BBOX Xmin': -107.62651, 'Unused0': 0, 'Unused1': 0, 'Unused2': 0, 'Unused3': 0, 'Unused4': 0, 'Version': 1000, 'BBOX Zmax': 0.0, 'Shape Type': 5, 'File Length': 260534}

    def test___iter__(self):
        shp = shp_file(pysal_examples.get_path('Point.shp'))
        points = [pt for pt in shp]
        expected = [{'Y': -0.25904661905760773, 'X': -0.00068176617532103578, 'Shape Type': 1},
                    {'Y': -0.25630328607387354, 'X': 0.11697145363360706,
                        'Shape Type': 1},
                    {'Y': -0.33930131004366804, 'X': 0.05043668122270728,
                        'Shape Type': 1},
                    {'Y': -0.41266375545851519, 'X': -0.041266375545851552,
                        'Shape Type': 1},
                    {'Y': -0.44017467248908293, 'X': -0.011462882096069604,
                        'Shape Type': 1},
                    {'Y': -0.46080786026200882, 'X': 0.027510917030567628,
                        'Shape Type': 1},
                    {'Y': -0.45851528384279472, 'X': 0.075655021834060809,
                        'Shape Type': 1},
                    {'Y': -0.43558951965065495, 'X': 0.11233624454148461,
                        'Shape Type': 1},
                    {'Y': -0.40578602620087334, 'X': 0.13984716157205224, 'Shape Type': 1}]
        assert points == expected

    def test___len__(self):
        shp = shp_file(pysal_examples.get_path('10740.shp'))
        assert len(shp) == 195

    def test_add_shape(self):
        shp = shp_file('test_point', 'w', 'POINT')
        points = [{'Shape Type': 1, 'X': 0, 'Y': 0},
                  {'Shape Type': 1, 'X': 1, 'Y': 1},
                  {'Shape Type': 1, 'X': 2, 'Y': 2},
                  {'Shape Type': 1, 'X': 3, 'Y': 3},
                  {'Shape Type': 1, 'X': 4, 'Y': 4}]
        for pt in points:
            shp.add_shape(pt)
        shp.close()

        for a, b in zip(points, shp_file('test_point')):
            self.assertEqual(a, b)
        os.remove('test_point.shp')
        os.remove('test_point.shx')

    def test_close(self):
        shp = shp_file(pysal_examples.get_path('10740.shp'))
        shp.close()
        self.assertEqual(shp.fileObj.closed, True)

    def test_get_shape(self):
        shp = shp_file(pysal_examples.get_path('Line.shp'))
        rec = shp.get_shape(0)
        expected = {'BBOX Ymax': -0.25832280562918325,
                    'NumPoints': 3,
                    'BBOX Ymin': -0.25895877033237352,
                    'NumParts': 1,
                    'Vertices': [(-0.0090539248870159517, -0.25832280562918325),
                                 (0.0074811573959305822, -0.25895877033237352),
                                 (
                                     0.0074811573959305822, -0.25895877033237352)],
                    'BBOX Xmax': 0.0074811573959305822,
                    'BBOX Xmin': -0.0090539248870159517,
                    'Shape Type': 3,
                    'Parts Index': [0]}
        self.assertEqual(expected, shp.get_shape(0))

    def test_next(self):
        shp = shp_file(pysal_examples.get_path('Point.shp'))
        points = [pt for pt in shp]
        expected = {'Y': -0.25904661905760773, 'X': -
                    0.00068176617532103578, 'Shape Type': 1}
        self.assertEqual(expected, next(shp))
        expected = {'Y': -0.25630328607387354, 'X':
                    0.11697145363360706, 'Shape Type': 1}
        self.assertEqual(expected, next(shp))

    def test_type(self):
        shp = shp_file(pysal_examples.get_path('Point.shp'))
        self.assertEqual("POINT", shp.type())
        shp = shp_file(pysal_examples.get_path('Polygon.shp'))
        self.assertEqual("POLYGON", shp.type())
        shp = shp_file(pysal_examples.get_path('Line.shp'))
        self.assertEqual("ARC", shp.type())


class test_shx_file(unittest.TestCase):
    def test___init__(self):
        shx = shx_file(pysal_examples.get_path('Point.shx'))
        assert isinstance(shx, shx_file)

    def test_add_record(self):
        shx = shx_file(pysal_examples.get_path('Point.shx'))
        expectedIndex = [(100, 20), (128, 20), (156, 20),
                         (184, 20), (212, 20), (240, 20),
                         (268, 20), (296, 20), (324, 20)]
        assert shx.index == expectedIndex
        shx2 = shx_file('test', 'w')
        for i, rec in enumerate(shx.index):
            id, location = shx2.add_record(rec[1])
            assert id == (i + 1)
            assert location == rec[0]
        assert shx2.index == shx.index
        shx2.close(shx._header)
        new_shx = open('test.shx', 'rb').read()
        expected_shx = open(pysal_examples.get_path('Point.shx'), 'rb').read()
        assert new_shx == expected_shx
        os.remove('test.shx')

    def test_close(self):
        shx = shx_file(pysal_examples.get_path('Point.shx'))
        shx.close(None)
        self.assertEqual(shx.fileObj.closed, True)


class TestNullShape(unittest.TestCase):
    def test_pack(self):
        null_shape = NullShape()
        self.assertEqual(b'\x00' * 4, null_shape.pack())

    def test_unpack(self):
        null_shape = NullShape()
        self.assertEqual(None, null_shape.unpack())


class TestPoint(unittest.TestCase):
    def test_pack(self):
        record = {"X": 5, "Y": 5, "Shape Type": 1}
        expected = b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14\x40\x00\x00\x00\x00\x00\x00\x14\x40"
        self.assertEqual(expected, Point.pack(record))

    def test_unpack(self):
        dat = bufferIO(b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14\x40\x00\x00\x00\x00\x00\x00\x14\x40")
        expected = {"X": 5, "Y": 5, "Shape Type": 1}
        self.assertEqual(expected, Point.unpack(dat))


class TestPolyLine(unittest.TestCase):
    def test_pack(self):
        record = {'BBOX Ymax': -0.25832280562918325, 'NumPoints': 3, 'BBOX Ymin': -0.25895877033237352, 'NumParts': 1, 'Vertices': [(-0.0090539248870159517, -0.25832280562918325), (0.0074811573959305822, -0.25895877033237352), (0.0074811573959305822, -0.25895877033237352)], 'BBOX Xmax': 0.0074811573959305822, 'BBOX Xmin': -0.0090539248870159517, 'Shape Type': 3, 'Parts Index': [0]}
        expected = b"""\x03\x00\x00\x00\xc0\x46\x52\x3a\xdd\x8a\x82\
\xbf\x3d\xc1\x65\xce\xc7\x92\xd0\xbf\x00\xc5\
\xa0\xe5\x8f\xa4\x7e\x3f\x6b\x40\x7f\x60\x5c\
\x88\xd0\xbf\x01\x00\x00\x00\x03\x00\x00\x00\
\x00\x00\x00\x00\xc0\x46\x52\x3a\xdd\x8a\x82\
\xbf\x6b\x40\x7f\x60\x5c\x88\xd0\xbf\x00\xc5\
\xa0\xe5\x8f\xa4\x7e\x3f\x3d\xc1\x65\xce\xc7\
\x92\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4\x7e\x3f\
\x3d\xc1\x65\xce\xc7\x92\xd0\xbf"""
        self.assertEqual(expected, PolyLine.pack(record))

    def test_unpack(self):
        dat = bufferIO(b"""\x03\x00\x00\x00\xc0\x46\x52\x3a\xdd\x8a\x82\
\xbf\x3d\xc1\x65\xce\xc7\x92\xd0\xbf\x00\xc5\
\xa0\xe5\x8f\xa4\x7e\x3f\x6b\x40\x7f\x60\x5c\
\x88\xd0\xbf\x01\x00\x00\x00\x03\x00\x00\x00\
\x00\x00\x00\x00\xc0\x46\x52\x3a\xdd\x8a\x82\
\xbf\x6b\x40\x7f\x60\x5c\x88\xd0\xbf\x00\xc5\
\xa0\xe5\x8f\xa4\x7e\x3f\x3d\xc1\x65\xce\xc7\
\x92\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4\x7e\x3f\
\x3d\xc1\x65\xce\xc7\x92\xd0\xbf""")
        expected = {'BBOX Ymax': -0.25832280562918325, 'NumPoints': 3, 'BBOX Ymin': -0.25895877033237352, 'NumParts': 1, 'Vertices': [(-0.0090539248870159517, -0.25832280562918325), (0.0074811573959305822, -0.25895877033237352), (0.0074811573959305822, -0.25895877033237352)], 'BBOX Xmax': 0.0074811573959305822, 'BBOX Xmin': -0.0090539248870159517, 'Shape Type': 3, 'Parts Index': [0]}
        self.assertEqual(expected, PolyLine.unpack(dat))


class TestMultiPoint(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, MultiPoint)


class TestPointZ(unittest.TestCase):
    def test_pack(self):
        record = {"X": 5, "Y": 5, "Z": 5, "M": 5, "Shape Type": 11}
        expected = b"\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00\x14@"
        self.assertEqual(expected, PointZ.pack(record))

    def test_unpack(self):
        dat = bufferIO(b"\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00\x14@")
        expected = {"X": 5, "Y": 5, "Z": 5, "M": 5, "Shape Type": 11}
        self.assertEqual(expected, PointZ.unpack(dat))


class TestPolyLineZ(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, PolyLineZ)


class TestPolyLineZ(unittest.TestCase):
    def test_pack(self):
        record = {'BBOX Ymax': -0.25832280562918325, 'NumPoints': 3, 'BBOX Ymin': -0.25895877033237352, 'NumParts': 1, 'Vertices': [(-0.0090539248870159517, -0.25832280562918325), (0.0074811573959305822, -0.25895877033237352), (0.0074811573959305822, -0.25895877033237352)], 'BBOX Xmax': 0.0074811573959305822, 'BBOX Xmin': -0.0090539248870159517, 'Shape Type': 13, 'Parts Index': [0], 'Zmin': 0, 'Zmax': 10, 'Zarray': [0, 5, 10], 'Mmin': 2, 'Mmax': 4, 'Marray': [2, 3, 4]}
        expected = b"""\r\x00\x00\x00\xc0FR:\xdd\x8a\x82\xbf=\xc1e\xce\xc7\x92\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4~?k@\x7f`\\\x88\xd0\xbf\x01\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\xc0FR:\xdd\x8a\x82\xbfk@\x7f`\\\x88\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4~?=\xc1e\xce\xc7\x92\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4~?=\xc1e\xce\xc7\x92\xd0\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00$@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00$@\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x10@\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x08@\x00\x00\x00\x00\x00\x00\x10@"""
        self.assertEqual(expected, PolyLineZ.pack(record))

    def test_unpack(self):
        dat = bufferIO(b"""\r\x00\x00\x00\xc0FR:\xdd\x8a\x82\xbf=\xc1e\xce\xc7\x92\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4~?k@\x7f`\\\x88\xd0\xbf\x01\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\xc0FR:\xdd\x8a\x82\xbfk@\x7f`\\\x88\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4~?=\xc1e\xce\xc7\x92\xd0\xbf\x00\xc5\xa0\xe5\x8f\xa4~?=\xc1e\xce\xc7\x92\xd0\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00$@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14@\x00\x00\x00\x00\x00\x00$@\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x10@\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x08@\x00\x00\x00\x00\x00\x00\x10@""")
        expected = {'BBOX Ymax': -0.25832280562918325, 'NumPoints': 3, 'BBOX Ymin': -0.25895877033237352, 'NumParts': 1, 'Vertices': [(-0.0090539248870159517, -0.25832280562918325), (0.0074811573959305822, -0.25895877033237352), (0.0074811573959305822, -0.25895877033237352)], 'BBOX Xmax': 0.0074811573959305822, 'BBOX Xmin': -0.0090539248870159517, 'Shape Type': 13, 'Parts Index': [0], 'Zmin': 0, 'Zmax': 10, 'Zarray': [0, 5, 10], 'Mmin': 2, 'Mmax': 4, 'Marray': [2, 3, 4]}
        self.assertEqual(expected, PolyLineZ.unpack(dat))


class TestPolygonZ(unittest.TestCase):
    def test_pack(self):
        record = {
            'BBOX Xmin': 0.0,
            'BBOX Xmax': 10.0,
            'BBOX Ymin': 0.0,
            'BBOX Ymax': 10.0,
            'NumPoints': 4,
            'NumParts': 1,
            'Vertices': [(0.0, 0.0),
                         (10.0, 10.0),
                         (10.0, 0.0),
                         (0.0, 0.0)],
            'Shape Type': 15,
            'Parts Index': [0],
            'Zmin': 0,
            'Zmax': 10,
            'Zarray': [0, 10, 0, 0],
            'Mmin': 2,
            'Mmax': 4,
            'Marray': [2, 4, 2, 2]
        }
        dat = bufferIO(PolygonZ.pack(record))
        self.assertEqual(record, PolygonZ.unpack(dat))


class TestMultiPointZ(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, MultiPointZ)
        # multi_point_z = MultiPointZ()


class TestPointM(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, PointM)
        # point_m = PointM()


class TestPolyLineM(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, PolyLineM)
        # poly_line_m = PolyLineM()


class TestPolygonM(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, PolygonM)
        # polygon_m = PolygonM()


class TestMultiPointM(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, MultiPointM)
        # multi_point_m = MultiPointM()


class TestMultiPatch(unittest.TestCase):
    def test___init__(self):
        self.assertRaises(NotImplementedError, MultiPatch)
        # multi_patch = MultiPatch()


class _TestPoints(unittest.TestCase):
    def test1(self):
        """ Test creating and reading Point Shape Files """
        shp = shp_file('test_point', 'w', 'POINT')
        points = [{'Shape Type': 1, 'X': 0, 'Y': 0}, {'Shape Type': 1, 'X': 1, 'Y': 1}, {'Shape Type': 1, 'X': 2, 'Y': 2}, {'Shape Type': 1, 'X': 3, 'Y': 3}, {'Shape Type': 1, 'X': 4, 'Y': 4}]
        for pt in points:
            shp.add_shape(pt)
        shp.close()

        shp = list(shp_file('test_point'))
        for a, b in zip(points, shp):
            self.assertEqual(a, b)
        os.remove('test_point.shp')
        os.remove('test_point.shx')


class _TestPolyLines(unittest.TestCase):
    def test1(self):
        """ Test creating and reading PolyLine Shape Files """
        lines = [[(0, 0), (4, 4)], [(1, 0), (5, 4)], [(2, 0), (6, 4)]]
        shapes = []
        for line in lines:
            x = [v[0] for v in line]
            y = [v[1] for v in line]
            rec = {}
            rec['BBOX Xmin'] = min(x)
            rec['BBOX Ymin'] = min(y)
            rec['BBOX Xmax'] = max(x)
            rec['BBOX Ymax'] = max(y)
            rec['NumPoints'] = len(line)
            rec['NumParts'] = 1
            rec['Vertices'] = line
            rec['Shape Type'] = 3
            rec['Parts Index'] = [0]
            shapes.append(rec)
        shp = shp_file('test_line', 'w', 'ARC')
        for line in shapes:
            shp.add_shape(line)
        shp.close()
        shp = list(shp_file('test_line'))
        for a, b in zip(shapes, shp):
            self.assertEqual(a, b)
        os.remove('test_line.shp')
        os.remove('test_line.shx')


class _TestPolygons(unittest.TestCase):
    def test1(self):
        """ Test creating and reading PolyLine Shape Files """
        lines = [[(0, 0), (4, 4), (5, 4), (
            1, 0), (0, 0)], [(1, 0), (5, 4), (6, 4), (2, 0), (1, 0)]]
        shapes = []
        for line in lines:
            x = [v[0] for v in line]
            y = [v[1] for v in line]
            rec = {}
            rec['BBOX Xmin'] = min(x)
            rec['BBOX Ymin'] = min(y)
            rec['BBOX Xmax'] = max(x)
            rec['BBOX Ymax'] = max(y)
            rec['NumPoints'] = len(line)
            rec['NumParts'] = 1
            rec['Vertices'] = line
            rec['Shape Type'] = 5
            rec['Parts Index'] = [0]
            shapes.append(rec)
        shp = shp_file('test_poly', 'w', 'POLYGON')
        for line in shapes:
            shp.add_shape(line)
        shp.close()
        shp = list(shp_file('test_poly'))
        for a, b in zip(shapes, shp):
            self.assertEqual(a, b)
        os.remove('test_poly.shp')
        os.remove('test_poly.shx')


if __name__ == '__main__':
    unittest.main()
